from __future__ import division, print_function, absolute_import, unicode_literals
#*****************************************************************
#    pyGSTi 0.9:  Copyright 2015 Sandia Corporation
#    This Software is released under the GPL license detailed
#    in the file "license.txt" in the top-level pyGSTi directory
#*****************************************************************
""" Text-parsering classes and functions to read input files."""

import os as _os
import sys as _sys
import numpy as _np
import warnings as _warnings
from scipy.linalg import expm as _expm
from collections import OrderedDict as _OrderedDict
import pyparsing as _pp

from .. import objects as _objs
from .. import tools as _tools

_pp.ParserElement.enablePackrat()
_sys.setrecursionlimit(10000)

class StdInputParser(object):
    """
    Encapsulates a text parser for reading GST input files.

    ** Grammar **

    expop   :: '^'
    multop  :: '*'
    integer :: '0'..'9'+
    real    :: ['+'|'-'] integer [ '.' integer [ 'e' ['+'|'-'] integer ] ]
    reflbl  :: (alpha | digit | '_')+

    nop     :: '{}'
    gate    :: 'G' [ lowercase | digit | '_' ]+
    strref  :: 'S' '[' reflbl ']'
    slcref  :: strref [ '[' integer ':' integer ']' ]
    expable :: gate | slcref | '(' string ')' | nop
    expdstr :: expable [ expop integer ]*
    string  :: expdstr [ [ multop ] expdstr ]*

    dataline :: string [ real ]+
    dictline :: reflbl string
    """

    def __init__(self):
        """ Creates a new StdInputParser object """

        def push_first( strg, loc, toks ):
            self.exprStack.append( toks[0] )
        def push_mult( strg, loc, toks ):
            self.exprStack.append( '*' )
        def push_slice( strg, loc, toks ):
            self.exprStack.append( 'SLICE' )
        #def push_count( strg, loc, toks ):
        #    self.exprStack.append( toks[0] )
        #    self.exprStack.append( 'COUNT' )

        # caps = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        # lowers = 'abcdefghijklmnopqrstuvwxyz' #caps.lower()
        digits = _pp.nums #"0123456789"  #same as "nums"
        #point = _pp.Literal( "." )
        #e     = _pp.CaselessLiteral( "E" )
        #real  = _pp.Combine( _pp.Word( "+-"+_pp.nums, _pp.nums ) +
        #                  _pp.Optional( point + _pp.Optional( _pp.Word( _pp.nums ) ) ) +
        #                  _pp.Optional( e + _pp.Word( "+-"+_pp.nums, _pp.nums ) ) ).setParseAction(push_first)
        # real = _pp.Regex(r'[+-]?(\d+(\.\d*)?|\.\d+)([eE][+-]?\d+)?') #faster than above
        nop   = _pp.Literal("{}").setParseAction(push_first)

        expop = _pp.Literal( "^" )
        lpar  = _pp.Literal( "(" ).suppress()
        rpar  = _pp.Literal( ")" ).suppress()
        # lbrk  = _pp.Literal( "[" ).suppress()
        # rbrk  = _pp.Literal( "]" ).suppress()

        integer = _pp.Word( digits ).setParseAction(push_first)
        reflbl  = _pp.Word(_pp.alphas+_pp.nums+"_").setParseAction(push_first)
        #gate   = _pp.Word( "G", lowers + digits + "_" ).setParseAction(push_first)
        gate    = _pp.Regex(r'G[a-z0-9_]+').setParseAction(push_first) #faster than above
        strref  = (_pp.Literal("S") + "[" + reflbl + "]" ).setParseAction(push_first)
        slcref  = (strref + _pp.Optional( ("[" + integer + ":" + integer + "]").setParseAction(push_slice)) )

        #bSimple = False #experimenting with possible parser speedups
        #if bSimple:
        #    string  = _pp.Forward()
        #    gateSeq = _pp.OneOrMore( gate )
        #    expable = (nop | gateSeq | lpar + gateSeq + rpar)
        #    expdstr = expable + _pp.Optional( (expop + integer).setParseAction(push_first) )
        #    string << expdstr + _pp.ZeroOrMore( (_pp.Optional("*") + expdstr).setParseAction(push_mult))
        #else:
        string  = _pp.Forward()
        expable = (gate | slcref | lpar + string + rpar | nop)
        expdstr = expable + _pp.ZeroOrMore( (expop + integer).setParseAction(push_first) )
        string << expdstr + _pp.ZeroOrMore( (_pp.Optional("*") + expdstr).setParseAction(push_mult)) #pylint: disable=expression-not-assigned

        #count = real.copy().setParseAction(push_count)
        #dataline = string + _pp.OneOrMore( count )
        dictline = reflbl + string

        self.string_parser = string
        #self.dataline_parser = dataline #OLD: when data lines had their own parser
        self.dictline_parser = dictline


    def _evaluateStack(self, s):
        op = s.pop()
        if op == "*":
            op2 = self._evaluateStack( s )
            op1 = self._evaluateStack( s )
            return op1 + op2 #tuple addition

        elif op == "^":
            exp = self._evaluateStack( s )
            op  = self._evaluateStack( s )
            return op*exp #tuple mulitplication = repeat op exp times

        elif op == "SLICE":
            upper = self._evaluateStack( s )
            lower = self._evaluateStack( s )
            op = self._evaluateStack( s )
            return op[lower:upper]

        #elif op == 'COUNT':
        #    cnt = float(s.pop())          # next item on stack is a count
        #    self.countList.insert(0, cnt) # so add it to countList and eval the rest of the stack
        #    return self._evaluateStack( s )

        elif op[0] == 'G':
            return (op,) #as tuple

        elif op[0] == 'S':
            reflabel = s.pop() # next item on stack is a reference label (keep as a str)
            return tuple(self.lookup[reflabel]) #lookup dict typically holds GateString objs...

        elif op == '{}': # no-op returns empty tuple
            return ()

        else:
            return int(op)

    def parse_gatestring(self, s, lookup={}):
        """
        Parse a gate string (string in grammar)

        Parameters
        ----------
        s : string
            The string to parse.

        lookup : dict, optional
            A dictionary with keys == reflbls and values == tuples of gate labels
            which can be used for substitutions using the S<reflbl> syntax.

        Returns
        -------
        tuple of gate labels
            Representing the gate string.
        """
        self.lookup = lookup
        self.exprStack = []
        try:
            self.string_parser.parseString(s)
        except _pp.ParseException as e:
            raise ValueError("Parsing error when parsing %s: %s" % (s,str(e)))
        #print "DB: result = ",result
        #print "DB: stack = ",self.exprStack
        return self._evaluateStack(self.exprStack)

    def parse_dataline(self, s, lookup={}, expectedCounts=-1):
        """
        Parse a data line (dataline in grammar)

        Parameters
        ----------
        s : string
            The string to parse.

        lookup : dict, optional
            A dictionary with keys == reflbls and values == tuples of gate labels
            which can be used for substitutions using the S<reflbl> syntax.

        expectedCounts : int, optional
            The expected number of counts to accompany the gate string on this
            data line.  If < 0, no check is performed; otherwise raises ValueError
            if the number of counts does not equal expectedCounts.

        Returns
        -------
        gateStringTuple : tuple
            The gate string as a tuple of gate labels.
        gateStringStr : string
            The gate string as represented as a string in the dataline
        counts : list
            List of counts following the gate string.
        """
        self.lookup = lookup
        self.exprStack = []
        self.countList = []

        #get counts from end of s
        parts = s.split(); counts = []
        for p in reversed(parts):
            try: f = float(p)
            except: break
            counts.append( f )
        counts.reverse() #because we appended them in reversed order
        totalCounts = len(counts) #in case expectedCounts is less
        if len(counts) > expectedCounts >= 0:
            counts = counts[0:expectedCounts]

        nCounts = len(counts)
        if expectedCounts >= 0 and nCounts != expectedCounts:
            raise ValueError("Found %d count columns when %d were expected" % (nCounts, expectedCounts))
        if nCounts == len(parts):
            raise ValueError("No gatestring column found -- all columns look like data")

        gateStringStr = " ".join(parts[0:len(parts)-totalCounts])
        gateStringTuple = self.parse_gatestring(gateStringStr, lookup)
        return gateStringTuple, gateStringStr, counts


    def parse_dictline(self, s):
        """
        Parse a gatestring dictionary line (dictline in grammar)

        Parameters
        ----------
        s : string
            The string to parse.

        Returns
        -------
        gateStringLabel : string
            The user-defined label to represent this gate string.
        gateStringTuple : tuple
            The gate string as a tuple of gate labels.
        gateStringStr : string
            The gate string as represented as a string in the dictline.
        """
        self.exprStack = []
        result = self.dictline_parser.parseString(s)
        #print "DB: result = ",result
        #print "DB: stack = ",self.exprStack
        gateStringLabel = result[0]
        gateStringTuple = self._evaluateStack(self.exprStack)
        gateStringStr = s[ s.index(gateStringLabel) + len(gateStringLabel) : ].strip()
        return gateStringLabel, gateStringTuple, gateStringStr

    def parse_stringfile(self, filename):
        """
        Parse a gatestring list file.

        Parameters
        ----------
        filename : string
            The file to parse.

        Returns
        -------
        list of GateStrings
            The gatestrings read from the file.
        """
        gatestring_list = [ ]
        with open(filename, 'r') as stringfile:
            for line in stringfile:
                line = line.strip()
                if len(line) == 0 or line[0] =='#': continue
                gatestring_list.append( _objs.GateString(self.parse_gatestring(line), line) )
        return gatestring_list

    def parse_dictfile(self, filename):
        """
        Parse a gatestring dictionary file.

        Parameters
        ----------
        filename : string
            The file to parse.

        Returns
        -------
        dict
           Dictionary with keys == gate string labels and values == GateStrings.
        """
        lookupDict = { }
        with open(filename, 'r') as dictfile:
            for line in dictfile:
                line = line.strip()
                if len(line) == 0 or line[0] =='#': continue
                label, tup, s = self.parse_dictline(line)
                lookupDict[ label ] = _objs.GateString(tup, s)
        return lookupDict

    def parse_datafile(self, filename, showProgress=True, collisionAction="aggregate"):
        """
        Parse a data set file into a DataSet object.

        Parameters
        ----------
        filename : string
            The file to parse.

        showProgress : bool, optional
            Whether or not progress should be displayed

        collisionAction : {"aggregate", "keepseparate"}
            Specifies how duplicate gate sequences should be handled.  "aggregate"
            adds duplicate-sequence counts, whereas "keepseparate" tags duplicate-
            sequence data with by appending a final "#<number>" gate label to the
            duplicated gate sequence.

        Returns
        -------
        DataSet
            A static DataSet object.
        """

        #Parse preamble -- lines beginning with # or ## until first non-# line
        preamble_directives = { }
        preamble_comments = []
        with open(filename, 'r') as datafile:
            for line in datafile:
                line = line.strip()
                if len(line) == 0 or line[0] != '#': break
                if line.startswith("## "):
                    parts = line[len("## "):].split("=")
                    if len(parts) == 2: # key = value
                        preamble_directives[ parts[0].strip() ] = parts[1].strip()
                elif line.startswith("#"):
                    preamble_comments.append(line[1:].strip())

        #Process premble
        orig_cwd = _os.getcwd()
        if len(_os.path.dirname(filename)) > 0: _os.chdir( _os.path.dirname(filename) ) #allow paths relative to datafile path
        try:
            if 'Lookup' in preamble_directives:
                lookupDict = self.parse_dictfile( preamble_directives['Lookup'] )
            else: lookupDict = { }
            if 'Columns' in preamble_directives:
                colLabels = [ l.strip() for l in preamble_directives['Columns'].split(",") ]
            else: colLabels = [ 'plus count', 'count total' ] #  spamLabel (' frequency' | ' count') | 'count total' |  ?? 'T0' | 'Tf' ??
            spamLabels,fillInfo = self._extractLabelsFromColLabels(colLabels)
            nDataCols = len(colLabels)
        finally:
            _os.chdir(orig_cwd)

        #Read data lines of data file
        dataset = _objs.DataSet(spamLabels=spamLabels,collisionAction=collisionAction,
                                comment="\n".join(preamble_comments))
        nLines  = 0
        with open(filename, 'r') as datafile:
            nLines = sum(1 for line in datafile)
        nSkip = int(nLines / 100.0)
        if nSkip == 0: nSkip = 1

        def is_interactive():
            import __main__ as main
            return not hasattr(main, '__file__')

        if is_interactive() and showProgress:
            try:
                import time
                from IPython.display import clear_output
                def display_progress(i,N):
                    time.sleep(0.001); clear_output()
                    print("Loading %s: %.0f%%" % (filename, 100.0*float(i)/float(N)))
                    _sys.stdout.flush()
            except:
                def display_progress(i,N): pass
        else:
            def display_progress(i,N): pass

        countDict = {}
        with open(filename, 'r') as inputfile:
            for (iLine,line) in enumerate(inputfile):
                if iLine % nSkip == 0 or iLine+1 == nLines: display_progress(iLine+1, nLines)

                line = line.strip()
                if len(line) == 0 or line[0] == '#': continue
                try:
                    gateStringTuple, gateStringStr, valueList = self.parse_dataline(line, lookupDict, nDataCols)
                except ValueError as e:
                    raise ValueError("%s Line %d: %s" % (filename, iLine, str(e)))

                self._fillDataCountDict( countDict, fillInfo, valueList )
                if all([ (abs(v) < 1e-9) for v in list(countDict.values())]):
                    _warnings.warn( "Dataline for gateString '%s' has zero counts and will be ignored" % gateStringStr)
                    continue #skip lines in dataset file with zero counts (no experiments done)
                dataset.add_count_dict(gateStringTuple, countDict) #Note: don't use gateStringStr since DataSet currently doesn't hold GateString objs (just tuples)

        dataset.done_adding_data()
        return dataset

    def _extractLabelsFromColLabels(self, colLabels ):
        spamLabels = []; countCols = []; freqCols = []; impliedCountTotCol1Q = -1
        for i,colLabel in enumerate(colLabels):
            if colLabel.endswith(' count'):
                spamLabel = colLabel[:-len(' count')]
                if spamLabel not in spamLabels: spamLabels.append( spamLabel )
                countCols.append( (spamLabel,i) )

            elif colLabel.endswith(' frequency'):
                if 'count total' not in colLabels:
                    raise ValueError("Frequency columns specified without count total")
                else: iTotal = colLabels.index( 'count total' )
                spamLabel = colLabel[:-len(' frequency')]
                if spamLabel not in spamLabels: spamLabels.append( spamLabel )
                freqCols.append( (spamLabel,i,iTotal) )

        if 'count total' in colLabels:
            if 'plus' in spamLabels and 'minus' not in spamLabels:
                spamLabels.append('minus')
                impliedCountTotCol1Q = colLabels.index( 'count total' )
            #TODO - add standard count completion for 2Qubit case?

        fillInfo = (countCols, freqCols, impliedCountTotCol1Q)
        return spamLabels, fillInfo


    def _fillDataCountDict(self, countDict, fillInfo, colValues):
        countCols, freqCols, impliedCountTotCol1Q = fillInfo

        for spamLabel,iCol in countCols:
            if colValues[iCol] > 0 and colValues[iCol] < 1:
                raise ValueError("Count column (%d) contains value(s) " % iCol +
                                 "between 0 and 1 - could this be a frequency?")
            countDict[spamLabel] = colValues[iCol]

        for spamLabel,iCol,iTotCol in freqCols:
            if colValues[iCol] < 0 or colValues[iCol] > 1.0:
                raise ValueError("Frequency column (%d) contains value(s) " % iCol +
                                 "outside of [0,1.0] interval - could this be a count?")
            countDict[spamLabel] = colValues[iCol] * colValues[iTotCol]

        if impliedCountTotCol1Q >= 0:
            countDict['minus'] = colValues[impliedCountTotCol1Q] - countDict['plus']
        #TODO - add standard count completion for 2Qubit case?
        return countDict


    def parse_multidatafile(self, filename, showProgress=True,
                            collisionAction="aggregate"):
        """
        Parse a multiple data set file into a MultiDataSet object.

        Parameters
        ----------
        filename : string
            The file to parse.

        showProgress : bool, optional
            Whether or not progress should be displayed

        collisionAction : {"aggregate", "keepseparate"}
            Specifies how duplicate gate sequences should be handled.  "aggregate"
            adds duplicate-sequence counts, whereas "keepseparate" tags duplicate-
            sequence data with by appending a final "#<number>" gate label to the
            duplicated gate sequence.

        Returns
        -------
        MultiDataSet
            A MultiDataSet object.
        """

        #Parse preamble -- lines beginning with # or ## until first non-# line
        preamble_directives = { }
        preamble_comments = []
        with open(filename, 'r') as multidatafile:
            for line in multidatafile:
                line = line.strip()
                if len(line) == 0 or line[0] != '#': break
                if line.startswith("## "):
                    parts = line[len("## "):].split("=")
                    if len(parts) == 2: # key = value
                        preamble_directives[ parts[0].strip() ] = parts[1].strip()
                elif line.startswith("#"):
                    preamble_comments.append(line[1:].strip())


        #Process premble
        orig_cwd = _os.getcwd()
        if len(_os.path.dirname(filename)) > 0:
            _os.chdir( _os.path.dirname(filename) ) #allow paths relative to datafile path
        try:
            if 'Lookup' in preamble_directives:
                lookupDict = self.parse_dictfile( preamble_directives['Lookup'] )
            else: lookupDict = { }
            if 'Columns' in preamble_directives:
                colLabels = [ l.strip() for l in preamble_directives['Columns'].split(",") ]
            else: colLabels = [ 'dataset1 plus count', 'dataset1 count total' ]
            dsSpamLabels, fillInfo = self._extractLabelsFromMultiDataColLabels(colLabels)
            nDataCols = len(colLabels)
        finally:
            _os.chdir(orig_cwd)

        #Read data lines of data file
        datasets = _OrderedDict()
        for dsLabel,spamLabels in dsSpamLabels.items():
            datasets[dsLabel] = _objs.DataSet(spamLabels=spamLabels,
                                              collisionAction=collisionAction)

        dsCountDicts = _OrderedDict()
        for dsLabel in dsSpamLabels: dsCountDicts[dsLabel] = {}

        nLines = 0
        with open(filename, 'r') as datafile:
            nLines = sum(1 for line in datafile)
        nSkip = max(int(nLines / 100.0),1)

        def is_interactive():
            import __main__ as main
            return not hasattr(main, '__file__')

        if is_interactive() and showProgress:
            try:
                import time
                from IPython.display import clear_output
                def display_progress(i,N):
                    time.sleep(0.001); clear_output()
                    print("Loading %s: %.0f%%" % (filename, 100.0*float(i)/float(N)))
                    _sys.stdout.flush()
            except:
                def display_progress(i,N): pass
        else:
            def display_progress(i,N): pass

        with open(filename, 'r') as inputfile:
            for (iLine,line) in enumerate(inputfile):
                if iLine % nSkip == 0 or iLine+1 == nLines: display_progress(iLine+1, nLines)

                line = line.strip()
                if len(line) == 0 or line[0] == '#': continue
                try:
                    gateStringTuple, _, valueList = self.parse_dataline(line, lookupDict, nDataCols)
                except ValueError as e:
                    raise ValueError("%s Line %d: %s" % (filename, iLine, str(e)))

                self._fillMultiDataCountDicts(dsCountDicts, fillInfo, valueList)
                for dsLabel, countDict in dsCountDicts.items():
                    datasets[dsLabel].add_count_dict(gateStringTuple, countDict)

        mds = _objs.MultiDataSet(comment="\n".join(preamble_comments))
        for dsLabel,ds in datasets.items():
            ds.done_adding_data()
            mds.add_dataset(dsLabel, ds)
        return mds


    #Note: spam labels must not contain spaces since we use spaces to separate
    # the spam label from the dataset label
    def _extractLabelsFromMultiDataColLabels(self, colLabels):
        dsSpamLabels = _OrderedDict()
        countCols = []; freqCols = []; impliedCounts1Q = []
        for i,colLabel in enumerate(colLabels):
            wordsInColLabel = colLabel.split() #split on whitespace into words
            if len(wordsInColLabel) < 3: continue #allow other columns we don't recognize

            if wordsInColLabel[-1] == 'count':
                spamLabel = wordsInColLabel[-2]
                dsLabel = wordsInColLabel[-3]
                if dsLabel not in dsSpamLabels:
                    dsSpamLabels[dsLabel] = [ spamLabel ]
                else: dsSpamLabels[dsLabel].append( spamLabel )
                countCols.append( (dsLabel,spamLabel,i) )

            elif wordsInColLabel[-1] == 'frequency':
                spamLabel = wordsInColLabel[-2]
                dsLabel = wordsInColLabel[-3]
                if '%s count total' % dsLabel not in colLabels:
                    raise ValueError("Frequency columns specified without" +
                                     "count total for dataset '%s'" % dsLabel)
                else: iTotal = colLabels.index( '%s count total' % dsLabel )

                if dsLabel not in dsSpamLabels:
                    dsSpamLabels[dsLabel] = [ spamLabel ]
                else: dsSpamLabels[dsLabel].append( spamLabel )
                freqCols.append( (dsLabel,spamLabel,i,iTotal) )

        for dsLabel,spamLabels in dsSpamLabels.items():
            if '%s count total' % dsLabel in colLabels:
                if 'plus' in spamLabels and 'minus' not in spamLabels:
                    dsSpamLabels[dsLabel].append('minus')
                    iTotal = colLabels.index( '%s count total' % dsLabel )
                    impliedCounts1Q.append( (dsLabel, iTotal) )
            #TODO - add standard count completion for 2Qubit case?

        fillInfo = (countCols, freqCols, impliedCounts1Q)
        return dsSpamLabels, fillInfo


    def _fillMultiDataCountDicts(self, countDicts, fillInfo, colValues):
        countCols, freqCols, impliedCounts1Q = fillInfo

        for dsLabel,spamLabel,iCol in countCols:
            if colValues[iCol] > 0 and colValues[iCol] < 1:
                raise ValueError("Count column (%d) contains value(s) " % iCol +
                                 "between 0 and 1 - could this be a frequency?")
            countDicts[dsLabel][spamLabel] = colValues[iCol]

        for dsLabel,spamLabel,iCol,iTotCol in freqCols:
            if colValues[iCol] < 0 or colValues[iCol] > 1.0:
                raise ValueError("Frequency column (%d) contains value(s) " % iCol +
                                 "outside of [0,1.0] interval - could this be a count?")
            countDicts[dsLabel][spamLabel] = colValues[iCol] * colValues[iTotCol]

        for dsLabel,iTotCol in impliedCounts1Q:
            countDicts[dsLabel]['minus'] = colValues[iTotCol] - countDicts[dsLabel]['plus']
        #TODO - add standard count completion for 2Qubit case?
        return countDicts



def _evalElement(el, bComplex):
    myLocal = { 'pi': _np.pi, 'sqrt': _np.sqrt }
    exec( "element = %s" % el, {"__builtins__": None}, myLocal )
    return complex( myLocal['element'] ) if bComplex else float( myLocal['element'] )

def _evalRowList(rows, bComplex):
    return _np.array( [ [ _evalElement(x,bComplex) for x in r ] for r in rows ],
                     'complex' if bComplex else 'd' )

def read_gateset(filename):
    """
    Parse a gateset file into a GateSet object.

    Parameters
    ----------
    filename : string
        The file to parse.

    Returns
    -------
    GateSet
    """

    def add_current_label():
        if cur_format == "StateVec":
            ar = _evalRowList( cur_rows, bComplex=True )
            if ar.shape == (1,2):
                spam_vecs[cur_label] = _tools.state_to_pauli_density_vec(ar[0,:])
            else: raise ValueError("Invalid state vector shape for %s: %s" % (cur_label,ar.shape))

        elif cur_format == "DensityMx":
            ar = _evalRowList( cur_rows, bComplex=True )
            if ar.shape == (2,2) or ar.shape == (4,4):
                spam_vecs[cur_label] = _tools.stdmx_to_ppvec(ar)
            else: raise ValueError("Invalid density matrix shape for %s: %s" % (cur_label,ar.shape))

        elif cur_format == "PauliVec":
            spam_vecs[cur_label] = _np.transpose( _evalRowList( cur_rows, bComplex=False ) )

        elif cur_format == "UnitaryMx":
            ar = _evalRowList( cur_rows, bComplex=True )
            if ar.shape == (2,2):
                gs.gates[cur_label] = _objs.FullyParameterizedGate(
                        _tools.unitary_to_pauligate_1q(ar))
            elif ar.shape == (4,4):
                gs.gates[cur_label] = _objs.FullyParameterizedGate(
                        _tools.unitary_to_pauligate_2q(ar))
            else: raise ValueError("Invalid unitary matrix shape for %s: %s" % (cur_label,ar.shape))

        elif cur_format == "UnitaryMxExp":
            ar = _evalRowList( cur_rows, bComplex=True )
            if ar.shape == (2,2):
                gs.gates[cur_label] = _objs.FullyParameterizedGate(
                        _tools.unitary_to_pauligate_1q( _expm(-1j * ar) ))
            elif ar.shape == (4,4):
                gs.gates[cur_label] = _objs.FullyParameterizedGate(
                        _tools.unitary_to_pauligate_2q( _expm(-1j * ar) ))
            else: raise ValueError("Invalid unitary matrix exponent shape for %s: %s" % (cur_label,ar.shape))

        elif cur_format == "PauliMx":
            gs.gates[cur_label] = _objs.FullyParameterizedGate( _evalRowList( cur_rows, bComplex=False ) )


    gs = _objs.GateSet()
    spam_vecs = _OrderedDict(); spam_labels = _OrderedDict(); remainder_spam_label = ""
    identity_vec = _np.transpose( _np.array( [ _np.sqrt(2.0), 0,0,0] ) )  #default = 1-QUBIT identity vector

    basis_abbrev = "pp" #default assumed basis
    basis_dims = None

    state = "look for label"
    cur_label = ""; cur_format = ""; cur_rows = []
    with open(filename) as inputfile:
        for line in inputfile:
            line = line.strip()

            if len(line) == 0:
                state = "look for label"
                if len(cur_label) > 0:
                    add_current_label()
                    cur_label = ""; cur_rows = []
                continue

            if line[0] == "#":
                continue

            if state == "look for label":
                if line.startswith("SPAMLABEL "):
                    eqParts = line[len("SPAMLABEL "):].split('=')
                    if len(eqParts) != 2: raise ValueError("Invalid spam label line: ", line)
                    if eqParts[1].strip() == "remainder":
                        remainder_spam_label = eqParts[0].strip()
                    else:
                        spam_labels[ eqParts[0].strip() ] = [ s.strip() for s in eqParts[1].split() ]

                elif line.startswith("IDENTITYVEC "):  #Vectorized form of identity density matrix in whatever basis is used
                    if line != "IDENTITYVEC None":  #special case for designating no identity vector, so default is not used
                        identity_vec  = _np.transpose( _evalRowList( [ line[len("IDENTITYVEC "):].split() ], bComplex=False ) )

                elif line.startswith("BASIS "): # Line of form "BASIS <abbrev> [<dims>]", where optional <dims> is comma-separated integers
                    parts = line[len("BASIS "):].split()
                    basis_abbrev = parts[0]
                    if len(parts) > 1:
                        basis_dims = list(map(int, "".join(parts[1:]).split(",")))
                        if len(basis_dims) == 1: basis_dims = basis_dims[0]
                    elif gs.get_dimension() is not None:
                        basis_dims = int(round(_np.sqrt(gs.get_dimension())))
                    elif len(spam_vecs) > 0:
                        basis_dims = int(round(_np.sqrt(list(spam_vecs.values())[0].size)))
                    else:
                        raise ValueError("BASIS directive without dimension, and cannot infer dimension!")
                else:
                    cur_label = line
                    state = "expect format"

            elif state == "expect format":
                cur_format = line
                if cur_format not in ["StateVec", "DensityMx", "UnitaryMx", "UnitaryMxExp", "PauliVec", "PauliMx"]:
                    raise ValueError("Expected object format for label %s and got line: %s -- must specify a valid object format" % (cur_label,line))
                state = "read object"

            elif state == "read object":
                cur_rows.append( line.split() )

    if len(cur_label) > 0:
        add_current_label()

    #Try to infer basis dimension if none is given
    if basis_dims is None:
        if gs.get_dimension() is not None:
            basis_dims = int(round(_np.sqrt(gs.get_dimension())))
        elif len(spam_vecs) > 0:
            basis_dims = int(round(_np.sqrt(list(spam_vecs.values())[0].size)))
        else:
            raise ValueError("Cannot infer basis dimension!")

    #Set basis
    gs.set_basis(basis_abbrev, basis_dims)

    #Default SPAMLABEL directive if none are give and rho and E vectors are:
    if len(spam_labels) == 0 and "rho" in spam_vecs and "E" in spam_vecs:
        spam_labels['plus'] = [ 'rho', 'E' ]
        spam_labels['minus'] = [ 'rho', 'remainder' ] #NEW default behavior
        # OLD default behavior: remainder_spam_label = 'minus'
    if len(spam_labels) == 0: raise ValueError("Must specify rho and E or spam labels directly.")

    #Make SPAMs
     #get unique rho and E names
    rho_names = list(_OrderedDict.fromkeys( [ rho for (rho,E) in list(spam_labels.values()) ] ) ) #if this fails, may be due to malformatted
    E_names   = list(_OrderedDict.fromkeys( [ E   for (rho,E) in list(spam_labels.values()) ] ) ) #  SPAMLABEL line (not 2 items to right of = sign)
    if "remainder" in rho_names:
        del rho_names[ rho_names.index("remainder") ]
    if "remainder" in E_names:
        del E_names[ E_names.index("remainder") ]

    #Order E_names and rho_names using spam_vecs ordering
    #rho_names = sorted(rho_names, key=spam_vecs.keys().index)
    #E_names = sorted(E_names, key=spam_vecs.keys().index)

     #add vectors to gateset
    for rho_nm in rho_names: gs.preps[rho_nm] = spam_vecs[rho_nm]
    for E_nm   in E_names:   gs.effects[E_nm] = spam_vecs[E_nm]

    gs.povm_identity = identity_vec

     #add spam labels to gateset
    for spam_label in spam_labels:
        (rho_nm,E_nm) = spam_labels[spam_label]
        gs.spamdefs[spam_label] = (rho_nm , E_nm)

    if len(remainder_spam_label) > 0:
        gs.spamdefs[remainder_spam_label] = ('remainder', 'remainder')

    return gs



from __future__ import division, print_function, absolute_import, unicode_literals
#*****************************************************************
#    pyGSTi 0.9:  Copyright 2015 Sandia Corporation
#    This Software is released under the GPL license detailed
#    in the file "license.txt" in the top-level pyGSTi directory
#*****************************************************************
""" Defines the GateString class and derived classes which represent gate strings."""

import numpy as _np

def _gateSeqToStr(seq):
    if len(seq) == 0: return "{}" #special case of empty gate string
    return ''.join(seq)

class GateString(object):
    """
    Encapsulates a gate string as a tuple of gate labels associated
    with a string representation for that tuple.

    Typically there are multiple string representations for the same tuple (for
    example "GxGx" and "Gx^2" both correspond to the tuple ("Gx","Gx") ) and it
    is convenient to store a specific string represntation along with the tuple.

    A GateString objects behaves very similarly to a tuple and most operations
    supported by a tuple are supported by a GateString (e.g. adding, hashing,
    testing for equality, indexing,  slicing, multiplying).
    """

    def __init__(self, tupleOfGateLabels, stringRepresentation=None, bCheck=True):
        """
        Create a new GateString object

        Parameters
        ----------
        tupleOfGateLabels : tuple or GateString (or None)
            A tuple of gate labels specifying the gate sequence, or None if the
            sequence should be obtained by evaluating stringRepresentation as
            a standard-text-format gate string (e.g. "GxGy", "Gx(Gy)^2, or "{}").

        stringRepresentation : string, optional
            A string representation of this GateString.

        bCheck : bool, optional
            If true, raise ValueEror if stringRepresentation does not evaluate
            to tupleOfGateLabels.
        """

        if tupleOfGateLabels is None and stringRepresentation is None:
            raise ValueError("tupleOfGateLabels and stringRepresentation cannot both be None");

        if tupleOfGateLabels is None or (bCheck and stringRepresentation is not None):
            from ..io import stdinput as _stdinput
            parser = _stdinput.StdInputParser()
            chkTuple = parser.parse_gatestring( stringRepresentation )
            if tupleOfGateLabels is None: tupleOfGateLabels = chkTuple
            elif tuple(tupleOfGateLabels) != chkTuple:
                raise ValueError("Error intializing GateString: " +
                            " tuple and string do not match: %s != %s"
                             % (tuple(tupleOfGateLabels),stringRepresentation))

        # if tupleOfGateLabels is a GateString, then copy it
        if isinstance(tupleOfGateLabels, GateString):
            self.tup = tupleOfGateLabels.tup
            if stringRepresentation is None:
                self.str = tupleOfGateLabels.str
            else:
                self.str = stringRepresentation

        else:
            if stringRepresentation is None:
                stringRepresentation = _gateSeqToStr( tupleOfGateLabels )

            self.tup = tuple(tupleOfGateLabels)
            self.str = str(stringRepresentation)

    #Conversion routines for evalTree usage -- TODO: make these member functions
    def to_pythonstr(self,gateLabels):
        """
        Convert this gate string into a python string, where each gate label is
        represented as a **single** character, starting with 'A' and contining
        down the alphabet.  This can be useful for processing gate strings
        using python's string tools (regex in particular).

        Parameters
        ----------
        gateLabels : tuple
           tuple containing all the gate labels that will be mapped to alphabet
           characters, beginning with 'A'.

        Returns
        -------
        string
            The converted gate string.

        Examples
        --------
            ('Gx','Gx','Gy','Gx') => "AABA"
        """
        assert(len(gateLabels) < 26) #Complain if we go beyond 'Z'
        translateDict = {}; c = 'A'
        for gateLabel in gateLabels:
            translateDict[gateLabel] = c
            c = chr(ord(c) + 1)
        return "".join([ translateDict[gateLabel] for gateLabel in self.tup ])

    @classmethod
    def from_pythonstr(cls,pythonString,gateLabels):
        """
        Create a GateString from a python string where each gate label is
        represented as a **single** character, starting with 'A' and contining
        down the alphabet.  This performs the inverse of to_pythonstr(...).

        Parameters
        ----------
        pythonString : string
            string whose individual characters correspond to the gate labels of a
            gate string.

        gateLabels : tuple
           tuple containing all the gate labels that will be mapped to alphabet
           characters, beginning with 'A'.

        Returns
        -------
        GateString

        Examples
        --------
            "AABA" => ('Gx','Gx','Gy','Gx')
        """
        assert(len(gateLabels) < 26) #Complain if we go beyond 'Z'
        translateDict = {}; c = 'A'
        for gateLabel in gateLabels:
            translateDict[c] = gateLabel
            c = chr(ord(c) + 1)
        return cls( tuple([ translateDict[c] for c in pythonString ]) )


    def __str__(self):
        return self.str

    def __len__(self):
        return len(self.tup)

    def __repr__(self):
        return "GateString(%s)" % self.str

    def __iter__(self):
        return self.tup.__iter__()

    def __add__(self,x):
        if not isinstance(x, GateString):
            raise ValueError("Can only add GateStrings objects to other GateString objects")
        if self.str != "{}":
            s = (self.str + x.str) if x.str != "{}" else self.str
        else: s = x.str
        return GateString(self.tup + x.tup, s, bCheck=False)

    def __mul__(self,x):
        assert( (isinstance(x,int) or _np.issubdtype(x,int)) and x >= 0)
        if x > 1: s = "(%s)^%d" % (self.str,x)
        elif x == 1: s = "(%s)" % self.str
        else: s = "{}"
        return GateString(self.tup * x, s, bCheck=False)

    def __pow__(self,x): #same as __mul__()
        return self.__mul__(x)

    def __eq__(self,x):
        if x is None: return False
        return self.tup == tuple(x) #better than x.tup since x can be a tuple

    def __lt__(self,x):
        return self.tup.__lt__(x)

    def __gt__(self,x):
        return self.tup.__gt__(x)

    def __hash__(self):
        return self.tup.__hash__()

    def __copy__(self):
        return GateString( self.tup, self.str, bCheck=False)

    #def __deepcopy__(self, memo):
    #    return GateString( self.tup, self.str, bCheck=False)

    def __getitem__(self, key):
        if isinstance( key, slice ):
            return GateString( self.tup.__getitem__(key) )
        return self.tup.__getitem__(key)

    def __setitem__(self, key, value):
        raise ValueError("Cannot set elements of GateString tuple (they're read-only)")



class WeightedGateString(GateString):
    """
    A GateString that contains an additional "weight" member used for
    building up weighted lists of gate strings.

    When two WeightedGateString objects are added together their weights
    add, and when a WeightedGateString object is multiplied by an integer
    (equivalent to being raised to a power) the weight is unchanged. When
    added to plain GateString objects, the plain GateString object is
    treated as having zero weight and the result is another WeightedGateString.
    """

    def __init__(self,tupleOfGateLabels, stringRepresentation=None, weight=1.0, bCheck=True):
        """
        Create a new WeightedGateString object

        Parameters
        ----------
        tupleOfGateLabels : tuple (or None)
            A tuple of gate labels specifying the gate sequence, or None if the
            sequence should be obtained by evaluating stringRepresentation as
            a standard-text-format gate string (e.g. "GxGy", "Gx(Gy)^2, or "{}").

        stringRepresentation : string, optional
            A string representation of this WeightedGateString.

        weight : float, optional
            the weight to assign this gate string.

        bCheck : bool, optional
            If true, raise ValueEror if stringRepresentation does not evaluate
            to tupleOfGateLabels.
        """
        self.weight = weight
        super(WeightedGateString,self).__init__(tupleOfGateLabels, stringRepresentation, bCheck)

    def __repr__(self):
        return "WeightedGateString(%s,%g)" % (self.str,self.weight)

    def __add__(self,x):
        tmp = super(WeightedGateString,self).__add__(x)
        x_weight = x.weight if type(x) == WeightedGateString else 0.0
        return WeightedGateString( tmp.tup, tmp.str, self.weight + x_weight, bCheck=False ) #add weights

    def __radd__(self,x):
        if isinstance(x, GateString):
            tmp = x.__add__(self)
            x_weight = x.weight if type(x) == WeightedGateString else 0.0
            return WeightedGateString( tmp.tup, tmp.str, x_weight + self.weight, bCheck=False )
        raise ValueError("Can only add GateStrings objects to other GateString objects")

    def __mul__(self,x):
        tmp = super(WeightedGateString,self).__mul__(x)
        return WeightedGateString(tmp.tup, tmp.str, self.weight, bCheck=False) #keep weight

    def __copy__(self):
        return WeightedGateString( self.tup, self.str, self.weight, bCheck=False )

#    def __deepcopy__(self, memo):
#        return WeightedGateString( self.tup, self.str, self.weight, bCheck=False )

    def __getitem__(self, key):
        if isinstance( key, slice ):
            return WeightedGateString( self.tup.__getitem__(key), None, self.weight, bCheck=False )
        return self.tup.__getitem__(key)


class CompressedGateString(object):
    """
    A "compressed" GateString class which reduces the memory or disk space
    required to hold the tuple part of a GateString by compressing it.

    One place where CompressedGateString objects can be useful is when saving
    large lists of long gate sequences in some non-human-readable format (e.g.
    pickle).  CompressedGateString objects *cannot* be used in place of
    GateString objects within pyGSTi, and so are *not* useful when manipulating
    and running algorithms which use gate sequences.
    """

    def __init__(self, gatestring, minLenToCompress=20, maxPeriodToLookFor=20):
        """
        Create a new CompressedGateString object

        Parameters
        ----------
        gatestring : GateString
            The gate string object which is compressed to create
            a new CompressedGateString object.

        minLenToCompress : int, optional
            The minimum length string to compress.  If len(gatestring)
            is less than this amount its tuple is returned.

        maxPeriodToLookFor : int, optional
            The maximum period length to use when searching for periodic
            structure within gatestring.  Larger values mean the method
            takes more time but could result in better compressing.
        """
        if not isinstance(gatestring, GateString):
            raise ValueError("CompressedGateStrings can only be created from existing GateString objects")
        self.tup = CompressedGateString.compress_gate_label_tuple(
            gatestring.tup, minLenToCompress, maxPeriodToLookFor)
        self.str = gatestring.str

    def expand(self):
        """
        Expands this compressed gate string into a GateString object.

        Returns
        -------
        GateString
        """
        tup = CompressedGateString.expand_gate_label_tuple(self.tup)
        return GateString(tup, self.str, bCheck=False)

    @staticmethod
    def _getNumPeriods(gateString, periodLen):
        n = 0
        if len(gateString) < periodLen: return 0
        while gateString[0:periodLen] == gateString[n*periodLen:(n+1)*periodLen]:
            n += 1
        return n


    @staticmethod
    def compress_gate_label_tuple(gateString, minLenToCompress=20, maxPeriodToLookFor=20):
        """
        Compress a gate string.  The result is tuple with a special compressed-
        gate-string form form that is not useable by other GST methods but is
        typically shorter (especially for long gate strings with a repetative
        structure) than the original gate string tuple.

        Parameters
        ----------
        gateString : tuple of gate labels or GateString
            The gate string to compress.

        minLenToCompress : int, optional
            The minimum length string to compress.  If len(gateString)
            is less than this amount its tuple is returned.

        maxPeriodToLookFor : int, optional
            The maximum period length to use when searching for periodic
            structure within gateString.  Larger values mean the method
            takes more time but could result in better compressing.

        Returns
        -------
        tuple
            The compressed (or raw) gate string tuple.
        """
        gateString = tuple(gateString) # converts from GateString or list to tuple if needed
        L = len(gateString)
        if L < minLenToCompress: return tuple(gateString)
        compressed = ["CCC"] #list for appending, then make into tuple at the end
        start = 0
        while start < L:
            #print "Start = ",start
            score = _np.zeros( maxPeriodToLookFor+1, 'd' )
            numperiods = _np.zeros( maxPeriodToLookFor+1, 'i' )
            for periodLen in range(1,maxPeriodToLookFor+1):
                n = CompressedGateString._getNumPeriods( gateString[start:], periodLen )
                if n == 0: score[periodLen] = 0
                elif n == 1: score[periodLen] = 4.1/periodLen
                else: score[periodLen] = _np.sqrt(periodLen)*n
                numperiods[periodLen] = n
            bestPeriodLen = _np.argmax(score)
            n = numperiods[bestPeriodLen]
            bestPeriod = gateString[start:start+bestPeriodLen]
            #print "Scores = ",score
            #print "num per = ",numperiods
            #print "best = %s ^ %d" % (str(bestPeriod), n)
            assert(n > 0 and bestPeriodLen > 0)
            if start > 0 and n == 1 and compressed[-1][1] == 1:
                compressed[-1] = (compressed[-1][0]+bestPeriod, 1)
            else:
                compressed.append( (bestPeriod, n) )
            start = start+bestPeriodLen*n

        return tuple(compressed)

    @staticmethod
    def expand_gate_label_tuple(compressedGateString):
        """
        Expand a compressed tuple created with compress_gate_label_tuple(...)
        into a tuple of gate labels.

        Parameters
        ----------
        compressedGateString : tuple
            a tuple in the compressed form created by
            compress_gate_label_tuple(...).

        Returns
        -------
        tuple
            A tuple of gate labels specifying the uncompressed gate string.
        """

        if len(compressedGateString) == 0: return ()
        if compressedGateString[0] != "CCC": return compressedGateString
        expandedString = []
        for (period,n) in compressedGateString[1:]:
            expandedString += period*n
        return tuple(expandedString)



#Now tested in unit tests
#if __name__ == "__main__":
#    wgs = WeightedGateString(('Gx',), weight=0.5)
#    gs = GateString(('Gx',) )
#    print ((gs + wgs)*2).weight
#    print (wgs + gs).weight



from __future__ import division, print_function, absolute_import, unicode_literals
import numbers as _numbers
from contextlib import contextmanager as _contextmanager
from copy       import deepcopy as _dc
import sys         as _sys
import math        as _math # used for digit formatting

def _num_digits(n):
    return int(_math.log10(n)) + 1 if n > 0 else 1

# This function isn't a part of the public interface, instead it has a wrapper in the VerbosityPrinter class
def _build_progress_bar (iteration, total, barLength = 100, numDecimals=2, fillChar='#',
                    emptyChar='-', prefix='Progress:', suffix='Complete', end='\n'):
    """
    Parameters
    ----------
    iteration   - int, required
      current iteration
    total       - int, required  :
      total iterations
    barLength   - int, optional  :
      character length of bar
    numDecimals - int, optional  :
      precision of progress percent
    fillChar    - str, optional  :
      replaces '#' as the bar-filling character
    emptyChar   - str, optional  :
      replaces '-' as the empty-bar character
    prefix      - str, optional  :
      message in front of the bar
    suffix      - str, optional  :
      message after the bar


    Returns
    -------
    formattedString - str:
      python string representing a progress bar
    """
    filledLength    = int(round(barLength * iteration / float(total)))
    percents        = round(100.00 * (iteration / float(total)), numDecimals)
    bar             = fillChar * filledLength + emptyChar * (barLength - filledLength)
    # Here, the \r (carriage return) is what replaces the last line that was printed
    carriageReturn  = end if iteration == total else '\r'
    formattedString = '%s [%s] %s%s %s%s' % (prefix, bar, percents, '%', suffix, carriageReturn)
    return formattedString

# Another hidden function for providing verbose progress output
def _build_verbose_iteration(iteration, total, prefix, suffix, end):
    digits = _num_digits(total)
    return '%s Iter %s of %s %s: %s' % (prefix, str(iteration+1).zfill(digits), total, suffix, end)

#############################################################################################################
#                                    The VerbosityPrinter Class itself                                      #
#############################################################################################################


# The class responsible for optionally logging output
class VerbosityPrinter():
    '''Class responsible for logging things to stdout or a file.

    Controls verbosity and can print progress bars. ex:

    >>> VerbosityPrinter(1)

    would construct a printer that printed out messages of level one or higher
    to the screen.

    >>> VerbosityPrinter(3, 'output.txt')

    would construct a printer that sends verbose output to a text file

    The static function :meth:`build_printer` will construct a printer from
    either an integer or an already existing printer.  it is a static method
    of the VerbosityPrinter class, so it is called like so:

    >>> VerbosityPrinter.build_printer(2)

    or

    >>> VerbostityPrinter.build_printer(VerbosityPrinter(3, 'output.txt'))

    ``printer.log('status')`` would log 'status' if the printers verbosity was
    one or higher. ``printer.log('status2', 2)`` would log 'status2' if the
    printer's verbosity was two or higher

    ``printer.error('something terrible happened')`` would ALWAYS log
    'something terrible happened'. ``printer.warning('something worrisome
    happened')`` would log if verbosity was one or higher - the same as a
    normal status.

    Both printer.error and printer.warning will prepend 'ERROR: ' or 'WARNING:
    ' to the message they are given.  Optionally, printer.log() can also
    prepend 'Status_n' to the message, where n is the message level.

    Logging of progress bars/iterations:

    >>> with printer_instance.progress_logging(verbosity):
    >>>     for i, item in enumerate(data):
    >>>         printer.show_progress(i, len(data))
    >>>         printer.log(...)

    will output either a progress bar or iteration statuses depending on the
    printer's verbosity
    '''

    # Rules for handling comm --This is a global variable-- (technically) it should probably only be set once, at the beginning of the program
    _commPath     = ''
    _commFileName = ''  # The name of the generated files, e.g. 'comm_output'. '' means don't output to comm files.  Must also be set
    _commFileExt  = '.txt'

    def _create_file(self, filename):
        with open(filename, 'w') as newFile:
            newFile.close()

    def _get_comm_file(self, comm_id):
        if len(VerbosityPrinter._commFileName) == 0: return ''
        return '%s%s%s%s' % (VerbosityPrinter._commPath, VerbosityPrinter._commFileName, comm_id, VerbosityPrinter._commFileExt)

    # The printer is initialized with a set verbosity, and an optional filename.
    # If a filename is not provided, VerbosityPrinter writes to stdout
    def __init__(self, verbosity, filename=None, comm=None, warnings=True):
        '''
        Customize a verbosity printer object

        Parameters
        ----------
        verbosity - int, required:
          How verbose the printer should be
        filename - string, optional:
          Where to put output (If none, output goes to screen)
        comm - mpi4py.MPI.Comm object, optional:
          Restricts output if the program is running in parallel
            ( By default, if the core is 0, output is sent to screen, and otherwise sent to commfiles 1, 2, and 3 (assuming 4 cores))
        warnings - bool, optional:
          don't print warnings
        '''
        if comm != None:
            if comm.Get_rank() != 0 and filename == None: # A filename will override the default comm behavior
                filename = self._get_comm_file(comm.Get_rank())
        self.verbosity = verbosity
        self.filename  = filename
        if filename is not None and len(filename) > 0:
            self._create_file(filename)
        self._comm            = comm
        self.progressLevel    = 0 # Used for queuing output while a progress bar is being shown
        self._delayQueue      = []
        self._progressStack   = []
        self._progressParamsStack   = []
        self.warnings         = warnings
        self.extra_indents    = 0 # Used for nested calls

    def clone(self):
        '''
        Instead of deepcopy, initialize a new printer object and feed it some select deepcopied members
        '''
        p = VerbosityPrinter(self.verbosity, self.filename, self._comm, self.warnings)
        p.progressLevel  = self.progressLevel
        p.extra_indents = self.extra_indents
        p._delayQueue    = _dc(self._delayQueue) # deepcopy
        p._progressStack = _dc(self._progressStack)
        p._progressParamsStack = _dc(self._progressParamsStack)
        return p

    # Function for converting between interfaces:
    # Accepts either a verbosity level (integer) or a pre-constructed VerbosityPrinter
    @staticmethod
    def build_printer(verbosity, comm=None):
        '''
        Function for converting between interfaces

        Parameters
        ----------
        verbosity : int or VerbosityPrinter object, required:
          object to build a printer from

        comm : mpi4py.MPI.Comm object, optional
          Comm object to build printers with. !Will override!

        Returns
        -------
        VerbosityPrinter:
          The printer object, constructed from either an integer or another printer

        '''
        if isinstance(verbosity, _numbers.Integral):
            printer = VerbosityPrinter(verbosity, comm=comm)
        else:
            printer = verbosity.clone() # deepcopy the printer object if it has been passed as a verbosity
            printer._comm = comm # override happens here
        return printer

    def __add__(self, other):
        '''
        Increase the verbosity of a VerbosityPrinter
        '''
        p = self.clone()
        p.verbosity += other
        p.extra_indents -= other
        return p

    def __sub__(self, other):
        '''
        Decrease the verbosity of a VerbosityPrinter
        '''
        p = self.clone()
        p.verbosity -= other
        p.extra_indents += other
        return p

    # Used once a file has been created - open the file whenever a message needs to be sent (rather than opening it for the entire program)
    def _append_to(self, filename, message):
        with open(filename, 'a') as output:
            output.write(message + '\n')

    # Hidden function for deciding what to do with our output
    def _put(self, message, flush=True, stderr=False):
        if self.filename is None: # Handles the case where comm is None or comm is rank 0
            if stderr:
                print(message, end='', file=_sys.stderr)
            else:
                print(message, end='')
            if flush:
                _sys.stdout.flush()
        elif len(self.filename) > 0:
            self._append_to(self.filename, message)

    # special function reserved for logging errors
    def error(self, message):
        '''
        Log an error to the screen/file

        Parameters
        ----------
        message : str
          the error message

        Returns
        -------
        None
        '''
        self._put('\nERROR: %s\n' % message, stderr=True)

    # special function reserved for logging warnings
    def warning(self, message):
        '''
        Log a warning to the screen/file if verbosity > 1

        Parameters
        ----------
        message : str
          the warning message

        Returns
        -------
        None
        '''
        if self.warnings:
            self._put('\nWARNING: %s\n' % message, stderr=True)

    def log(self, message, messageLevel=1, indentChar='  ', showStatustype=False, doIndent=True, indentOffset=0, end='\n', flush=True):
        '''
        Log a status message to screen/file
        Determines whether the message should be printed based on current verbosity setting,
        then sends the message to the appropriate output

        Parameters
        ----------
        message - str:
          Status message to be printed
        messageLevel - int, optional:
          Verbosity level required for the message to be shown
        indentChar - str, optional:
          Number of spaces to indent relative to verbosity
        showStatustype - bool, optional:
          output the status level of the message
        doIndent - bool, optional:
          do/dont indent the message
        indentOffset - int, optional:
          change indent based on verbosity
        end - str, optional:
          allows printing with no newlines etc
        flush - bool, optional:
          option to flush output

        Returns
        -------
        None
        '''
        if messageLevel <= self.verbosity:
            indent = (indentChar * (messageLevel-1+indentOffset 
                                    + self.extra_indents)) if doIndent else ''
               # messageLevel-1 so no indent at verbosity == 1
            statusType = 'Status Level %s:' % messageLevel if showStatustype else ''
            if end == '\n':
                #Special case where we process a message containing newlines
                formattedMessage = '\n'.join(['%s%s%s' % (indent, statusType, m) 
                                              for m in str(message).split('\n')]) + end
            else:
                formattedMessage = '%s%s%s%s' % (indent, statusType, message, end)

            if self.progressLevel > 0 and self.filename is None:
                self._delayQueue.append(indentChar + 'INVALID LEVEL: ' + formattedMessage)
            else:
                self._put(formattedMessage, flush=flush)

    def _progress_bar(self, iteration, total, barLength, numDecimals, fillChar, emptyChar, prefix, suffix, indent):
        progressBar = ''
        # 'self.progressLevel == 1' disallows nested progress bars !!!
        unnested = self.progressLevel == 1
        if unnested:
            progressBar =  _build_progress_bar(iteration, total, barLength, numDecimals,
                                                         fillChar, emptyChar, prefix, suffix)
            progressBar = indent + progressBar
        return progressBar

    def _verbose_iteration(self, iteration, total, prefix, suffix, verboseMessages, indent, end):
        iteration =  _build_verbose_iteration(iteration, total, prefix, suffix, end)
        iteration = indent + iteration
        for verboseMessage in verboseMessages:
            iteration += (indent + verboseMessage + '\n')
        return iteration

    def __str__(self):
        return 'Printer Object: Progress Level: %s Verbosity %s Indents %s' \
            % (self.progressLevel, self.verbosity, self.extra_indents)

    @_contextmanager
    def progress_logging(self, messageLevel=1):
        '''
        Context manager for logging progress bars/iterations
        (The printer will return to its normal, unrestricted state when the progress logging has finished)

        Parameters
        ----------
        messageLevel - int, optional:
          the verbosity level of the progressbar/set of iterations
        '''
        self._progressStack.append(messageLevel)
        self._progressParamsStack.append(None)
        if self.verbosity == messageLevel:
            self.progressLevel += 1
        yield
        self._end_progress()

    # A wrapper for show_progress that only works if verbosity is above a certain value (Status by default)
    def show_progress(self, iteration, total, messageLevel=1, barLength = 50, numDecimals=2, fillChar='#',
                    emptyChar='-', prefix='Progress:', suffix='', verboseMessages=[], indentChar='  ', end='\n'):

        """
        Parameters
        ----------
        iteration   - int, required
          current iteration
        total       - int, required  :
          total iterations
        barLength   - int, optional  :
          character length of bar
        numDecimals - int, optional  :
          precision of progress percent
        fillChar    - str, optional  :
          replaces '#' as the bar-filling character
        emptyChar   - str, optional  :
          replaces '-' as the empty-bar character
        prefix      - str, optional  :
          message in front of the bar
        suffix      - str, optional  :
          message after the bar
        verboseMessages - list(str), optional:
          list of messages to output alongside the iteration
        end - str, optional:
          String terminating the progress bar
        indentChar - str, optional:
          number of spaces to indent the progress bar
        """
        indent = indentChar * (self._progressStack[-1]-1 + self.extra_indents)
          # -1 so no indent at verbosity == 1

        # Print a standard progress bar if its verbosity matches ours,
        # Otherwise, Print verbose iterations if our verbosity is higher
        # Build either the progress bar or the verbose iteration status
        progress = ''
        if self.verbosity == self._progressStack[-1] and self.filename is None:
            progress = self._progress_bar(iteration, total, barLength, numDecimals,
                                          fillChar, emptyChar, prefix, suffix, indent)
            self._progressParamsStack[-1] = (iteration, total, barLength,
                                             numDecimals, fillChar, emptyChar,
                                             prefix, suffix, indent)
        elif self.verbosity >  self._progressStack[-1]:
            progress = self._verbose_iteration(iteration, total, prefix, suffix, verboseMessages, indent, end)

        self._put(progress) # send the progress logging to either file or stdout

    # must be explicitly called when the progress (e.g. loop) is done:
    #  This allows for early exits
    def _end_progress(self):
        if self.progressLevel > 0:
            last_progress_params = self._progressParamsStack.pop()
            if self.verbosity == self._progressStack.pop():
                if last_progress_params is not None:
                    (iteration, total, barLength,
                     numDecimals, fillChar, emptyChar,
                     prefix, suffix, indent) = last_progress_params
                    progress = self._progress_bar(iteration+1, total, barLength,
                                                  numDecimals, fillChar, emptyChar,
                                                  prefix, suffix, indent)
                    self._put(progress) # send the progress logging to either file or stdout

                # Show the statuses that were queued while the progressBar was active
                for item in self._delayQueue:
                    print(item)
                del self._delayQueue[:]
                self.progressLevel -= 1


########################################################################################################################################
#                                Demonstration of how the VerbosityPrinter class is used                                               #
########################################################################################################################################

# Some basic demonstrations of how to use the printer class with an arbitrary function

'''
if __name__ == "__main__":
    import threading
    import time

    def demo(verbosity):
        # usage of the show_progress function
        printer = VerbosityPrinter.build_printer(verbosity)
        data    = range(10)
        with printer.progress_logging(2):
            for i, item in enumerate(data):
                printer.show_progress(i, len(data)-1, messageLevel=2,
                          verboseMessages=['%s gates' % i], prefix='--- GST (', suffix=') ---')
                time.sleep(.05)

    def nested_demo(verbosity):
        printer = VerbosityPrinter.build_printer(verbosity)
        printer.warning('Beginning demonstration of the verbosityprinter class. This could go wrong..')
        data    = range(10)
        with printer.progress_logging(1):
            for i, item in enumerate(data):
                printer.show_progress(i, len(data)-1, messageLevel=1,
                      verboseMessages=['%s gate strings' % i], prefix='-- IterativeGST (', suffix=') --')
                if i == 5:
                    printer.error('The iterator is five. This caused an error, apparently')
                demo(printer - 1)

    print('\nTersest: \n')
    nested_demo(0)

    print('\nTerse: \n')
    nested_demo(1)

    print('\nStandard: \n')
    nested_demo(2)

    print('\nVerbose: \n')
    nested_demo(3)

    print('\nMost Verbose: \n')
    nested_demo(4)
    # Create four threads of different verbosities, each of which write output to their own file

    threads = []
    for i in range(4):
        # Each thread is started with a printer that is assigned a different verbosity
        printer = VerbosityPrinter(i, 'output%s.txt' % i)
        t = threading.Thread(target=demo, args = [printer])
        t.daemon = True
        t.start()
        threads.append(t)

    for thread in threads:
        thread.join()
'''



from __future__ import division, print_function, absolute_import, unicode_literals
#*****************************************************************
#    pyGSTi 0.9:  Copyright 2015 Sandia Corporation
#    This Software is released under the GPL license detailed
#    in the file "license.txt" in the top-level pyGSTi directory
#*****************************************************************
"""Functions related to computation of the log-likelihood."""

import numpy as _np
import warnings as _warnings
import itertools as _itertools
#import time as _time
from . import basistools as _bt
from . import jamiolkowski as _jam
from . import mpitools as _mpit
from . import slicetools as _slct

TOL = 1e-20


# Functions for computing the log liklihood function and its derivatives

def create_count_vec_dict(spamLabels, dataset, gatestring_list):
    """
    Create a count-vector dictionary that is useful for speeding up multiple
    evaluations of logl(...).   The returned dictionary has keys that are
    spam labels and values that are numpy vectors containing the dataset counts
    for that spam label for each gate string in gatestring_list.

    Parameters
    ----------
    spamLabels : list of strings
        List of the spam labels to include as keys in returned dict.

    dataset : DataSet
        The dataset to extract counts from.

    gatestring_list : list of (tuples or GateStrings)
        List of the gate strings to extract counts for, which
        determines the ordering of the counts within each dictionary
        value.

    Returns
    -------
    dict
        as described above.
    """
    countVecDict = { }
    for spamLabel in spamLabels:
        countVecDict[spamLabel] = _np.array( [ dataset[gs][spamLabel] for gs in gatestring_list ] )
    return countVecDict


def fill_count_vecs(mxToFill, spam_label_rows, dataset, gatestring_list):
    """
    Fill a matrix of counts that is useful for speeding up multiple
    evaluations of logl(...).   Identical to create_count_vec_dict except
    counts for a given spam label are placed into a row of mxToFill
    instead of into a returned dictionary.

    Parameters
    ----------
    mxToFill : numpy ndarray
        an already-allocated KxS numpy array, where K is larger
        than the maximum value in spam_label_rows and S is equal
        to the number of gate strings (lenght of gatestring_list).

    spam_label_rows : dictionary
        a dictionary with keys == spam labels and values which
        are integer row indices into mxToFill, specifying the
        correspondence between rows of mxToFill and spam labels.

    dataset : DataSet
        The dataset to extract counts from.

    gatestring_list : list of (tuples or GateStrings)
        List of the gate strings to extract counts for, which
        determines the ordering of the counts within each dictionary
        value.

    Returns
    -------
    None
    """
    for spamLabel,iRow in spam_label_rows.items():
        mxToFill[iRow,:] = [ dataset[gs][spamLabel] for gs in gatestring_list ]




 # The log(Likelihood) within the standard (non-Poisson) picture is:
 #
 # L = prod_{i,sl} p_{i,sl}^N_{i,sl}
 #
 # Where i indexes the gate string, and sl indexes the spam label.  N[i] is the total counts
 #  for the i-th gatestring, and so sum_{sl} N_{i,sl} == N[i]. We can take the log:
 #
 # log L = sum_{i,sl} N_{i,sl} log(p_{i,sl})
 #
 #   after patching (linear extrapolation below min_p and ignore f == 0 terms ( 0*log(0) == 0 ) ):
 #
 # logl = sum_{i,sl} N_{i,sl} log(p_{i,sl})                                                        if p_{i,sl} >= min_p and N_{i,sl} > 0
 #                   N_{i,sl} log(min_p)     + S * (p_{i,sl} - min_p) + S2 * (p_{i,sl} - min_p)**2 if p_{i,sl} < p_min and N_{i,sl} > 0
 #                   0                                                                             if N_{i,sl} == 0
 #
 # dlogL = sum_{i,sl} N_{i,sl} / p_{i,sl} * dp                    if p_{i,sl} >= min_p and N_{i,sl} > 0
 #                    (S + 2*S2*(p_{i,sl} - min_p)) * dp          if p_{i,sl} < p_min and N_{i,sl} > 0
 #                    0                                           if N_{i,sl} == 0
 #
 # hlogL = sum_{i,sl} -N_{i,sl} / p_{i,sl}**2 * dp1 * dp2 +  N_{i,sl} / p_{i,sl} *hp        if p_{i,sl} >= min_p and N_{i,sl} > 0
 #                    2*S2* dp1 * dp2 + (S + 2*S2*(p_{i,sl} - min_p)) * hp                  if p_{i,sl} < p_min and N_{i,sl} > 0
 #                    0                                                                     if N_{i,sl} == 0
 #
 #  where S = N_{i,sl} / min_p is the slope of the line tangent to logl at min_p
 #    and S2 = 0.5*( -N_{i,sl} / min_p**2 ) is 1/2 the 2nd derivative of the logl term at min_p
 #   and hlogL == d/d1 ( d/d2 ( logl ) )  -- i.e. dp2 is the *first* derivative performed...


 #Note: Poisson picture entered use when we allowed an EVec which was 1-{other EVecs} -- a
 # (0,-1) spam index -- instead of assuming all probabilities of a given gat string summed
 # to one -- a (-1,-1) spam index.  The poisson picture gives a correct log-likelihood
 # description when the probabilities (for a given gate string) may not sum to one, by
 # interpreting them each as rates.  In the standard picture, large gatestring probabilities
 # are not penalized (each standard logL term increases monotonically with each probability,
 # and the reason this is ok when the probabilities sum to one is that for a probabilility
 # that gets close to 1, there's another that is close to zero, and logL is very negative
 # near zero.

 # The log(Likelihood) within the Poisson picture is:
 #
 # L = prod_{i,sl} lambda_{i,sl}^N_{i,sl} e^{-lambda_{i,sl}} / N_{i,sl}!
 #
 # Where lamba_{i,sl} := p_{i,sl}*N[i] is a rate, i indexes the gate string,
 #  and sl indexes the spam label.  N[i] is the total counts for the i-th gatestring, and
 #  so sum_{sl} N_{i,sl} == N[i]. We can ignore the p-independent N_j! and take the log:
 #
 # log L = sum_{i,sl} N_{i,sl} log(N[i]*p_{i,sl}) - N[i]*p_{i,sl}
 #       = sum_{i,sl} N_{i,sl} log(p_{i,sl}) - N[i]*p_{i,sl}   (where we ignore the p-independent log(N[i]) terms)
 #
 #   after patching (linear extrapolation below min_p and "softening" f == 0 terms w/cubit below radius "a"):
 #
 # logl = sum_{i,sl} N_{i,sl} log(p_{i,sl}) - N[i]*p_{i,sl}                                                        if p_{i,sl} >= min_p and N_{i,sl} > 0
 #                   N_{i,sl} log(min_p)    - N[i]*min_p    + S * (p_{i,sl} - min_p) + S2 * (p_{i,sl} - min_p)**2  if p_{i,sl} < p_min and N_{i,sl} > 0
 #                   0                      - N[i]*p_{i,sl}                                                        if N_{i,sl} == 0 and p_{i,sl} >= a
 #                   0                      - N[i]*( -(1/(3a**2))p_{i,sl}**3 + p_{i,sl}**2/a + (1/3)*a )           if N_{i,sl} == 0 and p_{i,sl} < a
 #
 # dlogL = sum_{i,sl} [ N_{i,sl} / p_{i,sl} - N[i] ] * dp                   if p_{i,sl} >= min_p and N_{i,sl} > 0
 #                    (S + 2*S2*(p_{i,sl} - min_p)) * dp                    if p_{i,sl} < p_min and N_{i,sl} > 0
 #                    -N[i] * dp                                            if N_{i,sl} == 0 and p_{i,sl} >= a
 #                    -N[i] * ( (-1/a**2)p_{i,sl}**2 + 2*p_{i,sl}/a ) * dp  if N_{i,sl} == 0 and p_{i,sl} < a
 #
 # hlogL = sum_{i,sl} -N_{i,sl} / p_{i,sl}**2 * dp1 * dp2 + [ N_{i,sl} / p_{i,sl} - N[i] ]*hp      if p_{i,sl} >= min_p and N_{i,sl} > 0
 #                    2*S2* dp1 * dp2 + (S + 2*S2*(p_{i,sl} - min_p)) * hp                         if p_{i,sl} < p_min and N_{i,sl} > 0
 #                    -N[i] * hp                                                                   if N_{i,sl} == 0 and p_{i,sl} >= a
 #                    -N[i]*( (-2/a**2)p_{i,sl} + 2/a ) * dp1 * dp2
 #                        - N[i]*( (-1/a**2)p_{i,sl}**2 + 2*p_{i,sl}/a ) * hp                      if N_{i,sl} == 0 and p_{i,sl} < a
 #
 #  where S = N_{i,sl} / min_p - N[i] is the slope of the line tangent to logl at min_p
 #    and S2 = 0.5*( -N_{i,sl} / min_p**2 ) is 1/2 the 2nd derivative of the logl term at min_p so
 #    logL_term = logL_term(min_p) + S * (p-min_p) + S2 * (p-min_p)**2
 #   and hlogL == d/d1 ( d/d2 ( logl ) )  -- i.e. dp2 is the *first* derivative performed...
 #
 # For cubit interpolation, use function F(p) (derived by Robin: match value, 1st-deriv, 2nd-deriv at p == r, and require min at p == 0):
 #  Given a radius r << 1 (but r>0):
 #   F(p) = piecewise{ if( p>r ) then p; else -(1/3)*p^3/r^2 + p^2/r + (1/3)*r }
 #  OLD: quadratic that doesn't match 2nd-deriv:
 #   F(p) = piecewise{ if( p>r ) then p; else (r-p)^2/(2*r) + p }




def logl(gateset, dataset, gatestring_list=None,
         minProbClip=1e-6, probClipInterval=(-1e6,1e6), radius=1e-4,
         evalTree=None, countVecMx=None, poissonPicture=True, check=False):
    """
    The log-likelihood function.

    Parameters
    ----------
    gateset : GateSet
        Gateset of parameterized gates

    dataset : DataSet
        Probability data

    gatestring_list : list of (tuples or GateStrings), optional
        Each element specifies a gate string to include in the log-likelihood
        sum.  Default value of None implies all the gate strings in dataset
        should be used.

    minProbClip : float, optional
        The minimum probability treated normally in the evaluation of the log-likelihood.
        A penalty function replaces the true log-likelihood for probabilities that lie
        below this threshold so that the log-likelihood never becomes undefined (which improves
        optimizer performance).

    probClipInterval : 2-tuple or None, optional
        (min,max) values used to clip the probabilities predicted by gatesets during MLEGST's
        search for an optimal gateset (if not None).  if None, no clipping is performed.

    radius : float, optional
        Specifies the severity of rounding used to "patch" the zero-frequency
        terms of the log-likelihood.

    evalTree : evaluation tree, optional
      given by a prior call to bulk_evaltree for the same gatestring_list.
      Significantly speeds up evaluation of log-likelihood, even more so
      when accompanied by countVecMx (see below).

    countVecMx : numpy array, optional
      Two-dimensional numpy array whose rows correspond to the gate's spam
      labels (i.e. gateset.get_spam_labels()).  Each row is  contains the
      dataset counts for that spam label for each gate string in gatestring_list.
      Use fill_count_vecs(...) to generate this quantity once for multiple
      evaluations of the log-likelihood function which use the same dataset.

    poissonPicture : boolean, optional
        Whether the log-likelihood-in-the-Poisson-picture terms should be included
        in the returned logl value.

    check : boolean, optional
      If True, perform extra checks within code to verify correctness.  Used
      for testing, and runs much slower when True.

    Returns
    -------
    float
        The log likelihood
    """

    if gatestring_list is None:
        gatestring_list = list(dataset.keys())

    spamLabels = gateset.get_spam_labels() #this list fixes the ordering of the spam labels
    spam_lbl_rows = { sl:i for (i,sl) in enumerate(spamLabels) }

    probs = _np.empty( (len(spamLabels),len(gatestring_list)), 'd' )
    if countVecMx is None:
        countVecMx = _np.empty( (len(spamLabels),len(gatestring_list)), 'd' )
        fill_count_vecs(countVecMx, spam_lbl_rows, dataset, gatestring_list)

    totalCntVec = _np.sum(countVecMx, axis=0)

    #freqs = countVecMx / totalCntVec[None,:]
    #freqs_nozeros = _np.where(countVecMx == 0, 1.0, freqs) # set zero freqs to 1.0 so np.log doesn't complain
    #freqTerm = countVecMx * ( _np.log(freqs_nozeros) - 1.0 )
    #freqTerm[ countVecMx == 0 ] = 0.0 # set 0 * log(0) terms explicitly to zero since numpy doesn't know this limiting behavior

    a = radius # parameterizes "roundness" of f == 0 terms
    min_p = minProbClip

    if evalTree is None:
        evalTree = gateset.bulk_evaltree(gatestring_list)

    gateset.bulk_fill_probs(probs, spam_lbl_rows, evalTree, probClipInterval, check)
    pos_probs = _np.where(probs < min_p, min_p, probs)

    if poissonPicture:
        S = countVecMx / min_p - totalCntVec[None,:] # slope term that is derivative of logl at min_p
        S2 = -0.5 * countVecMx / (min_p**2)          # 2nd derivative of logl term at min_p
        v = countVecMx * _np.log(pos_probs) - totalCntVec[None,:]*pos_probs # dims K x M (K = nSpamLabels, M = nGateStrings)
        v = _np.minimum(v,0)  #remove small positive elements due to roundoff error (above expression *cannot* really be positive)
        v = _np.where( probs < min_p, v + S*(probs - min_p) + S2*(probs - min_p)**2, v) #quadratic extrapolation of logl at min_p for probabilities < min_p
        v = _np.where( countVecMx == 0, -totalCntVec[None,:] * _np.where(probs >= a, probs, (-1.0/(3*a**2))*probs**3 + probs**2/a + a/3.0), v)
           #special handling for f == 0 poissonPicture terms using quadratic rounding of function with minimum: max(0,(a-p))^2/(2a) + p

    else: #(the non-poisson picture requires that the probabilities of the spam labels for a given string are constrained to sum to 1)
        S = countVecMx / min_p               # slope term that is derivative of logl at min_p
        S2 = -0.5 * countVecMx / (min_p**2)  # 2nd derivative of logl term at min_p
        v = countVecMx * _np.log(pos_probs) # dims K x M (K = nSpamLabels, M = nGateStrings)
        v = _np.minimum(v,0)  #remove small positive elements due to roundoff error (above expression *cannot* really be positive)
        v = _np.where( probs < min_p, v + S*(probs - min_p) + S2*(probs - min_p)**2, v) #quadratic extrapolation of logl at min_p for probabilities < min_p
        v = _np.where( countVecMx == 0, 0.0, v)

    #DEBUG
    #print "num clipped = ",_np.sum(probs < min_p)," of ",probs.shape
    #print "min/max probs = ",min(probs.flatten()),",",max(probs.flatten())
    #for i in range(v.shape[1]):
    #    print "%d %.0f (%f) %.0f (%g)" % (i,v[0,i],probs[0,i],v[1,i],probs[1,i])

    # v[iSpamLabel,iGateString] contains all logl contributions
    return _np.sum(v) # sum over *all* dimensions




def logl_jacobian(gateset, dataset, gatestring_list=None,
                  minProbClip=1e-6, probClipInterval=(-1e6,1e6), radius=1e-4,
                  evalTree=None, countVecMx=None, poissonPicture=True, check=False):
    """
    The jacobian of the log-likelihood function.

    Parameters
    ----------
    gateset : GateSet
        Gateset of parameterized gates (including SPAM)

    dataset : DataSet
        Probability data

    gatestring_list : list of (tuples or GateStrings), optional
        Each element specifies a gate string to include in the log-likelihood
        sum.  Default value of None implies all the gate strings in dataset
        should be used.

    minProbClip : float, optional
        The minimum probability treated normally in the evaluation of the log-likelihood.
        A penalty function replaces the true log-likelihood for probabilities that lie
        below this threshold so that the log-likelihood never becomes undefined (which improves
        optimizer performance).

    probClipInterval : 2-tuple or None, optional
        (min,max) values used to clip the probabilities predicted by gatesets during MLEGST's
        search for an optimal gateset (if not None).  if None, no clipping is performed.

    radius : float, optional
        Specifies the severity of rounding used to "patch" the zero-frequency
        terms of the log-likelihood.

    evalTree : evaluation tree, optional
        given by a prior call to bulk_evaltree for the same gatestring_list.
        Significantly speeds up evaluation of log-likelihood derivatives, even
        more so when accompanied by countVecMx (see below).  Defaults to None.

    countVecMx : numpy array, optional
      Two-dimensional numpy array whose rows correspond to the gate's spam
      labels (i.e. gateset.get_spam_labels()).  Each row is  contains the
      dataset counts for that spam label for each gate string in gatestring_list.
      Use fill_count_vecs(...) to generate this quantity once for multiple
      evaluations of the log-likelihood function which use the same dataset.

    poissonPicture : boolean, optional
        Whether the Poisson-picutre log-likelihood should be differentiated.

    check : boolean, optional
        If True, perform extra checks within code to verify correctness.  Used
        for testing, and runs much slower when True.

    Returns
    -------
    numpy array
      array of shape (M,), where M is the length of the vectorized gateset.
    """

    nP = gateset.num_params()
    jac = _np.zeros([1,nP])

    if gatestring_list is None:
        gatestring_list = list(dataset.keys())

    spamLabels = gateset.get_spam_labels() #this list fixes the ordering of the spam labels
    spam_lbl_rows = { sl:i for (i,sl) in enumerate(spamLabels) }

    if countVecMx is None:
        countVecMx = _np.empty( (len(spamLabels),len(gatestring_list)), 'd' )
        fill_count_vecs(countVecMx, spam_lbl_rows, dataset, gatestring_list)

    probs = _np.empty( (len(spamLabels),len(gatestring_list)), 'd' )
    dprobs = _np.empty( (len(spamLabels),len(gatestring_list),nP), 'd' )
    totalCntVec = _np.sum(countVecMx, axis=0)

    #freqs = cntVecMx / totalCntVec[None,:]
    #freqs_nozeros = _np.where(cntVecMx == 0, 1.0, freqs) # set zero freqs to 1.0 so np.log doesn't complain
    #freqTerm = cntVecMx * ( _np.log(freqs_nozeros) - 1.0 )
    #freqTerm[ cntVecMx == 0 ] = 0.0 # set 0 * log(0) terms explicitly to zero since numpy doesn't know this limiting behavior
    #minusCntVecMx = -1.0 * cntVecMx

    a = radius # parameterizes "roundness" of f == 0 terms
    min_p = minProbClip

    if evalTree is None:
        evalTree = gateset.bulk_evaltree(gatestring_list)

    gateset.bulk_fill_dprobs(dprobs, spam_lbl_rows, evalTree,
                            prMxToFill=probs, clipTo=probClipInterval, check=check)

    pos_probs = _np.where(probs < min_p, min_p, probs)

    if poissonPicture:
        S = countVecMx / min_p - totalCntVec[None,:] # slope term that is derivative of logl at min_p
        S2 = -0.5 * countVecMx / (min_p**2)          # 2nd derivative of logl term at min_p
        v = countVecMx * _np.log(pos_probs) - totalCntVec[None,:]*pos_probs # dims K x M (K = nSpamLabels, M = nGateStrings)
        v = _np.minimum(v,0)  #remove small positive elements due to roundoff error (above expression *cannot* really be positive)
        v = _np.where( probs < min_p, v + S*(probs - min_p) + S2*(probs - min_p)**2, v) #quadratic extrapolation of logl at min_p for probabilities < min_p
        v = _np.where( countVecMx == 0, -totalCntVec[None,:] * _np.where(probs >= a, probs, (-1.0/(3*a**2))*probs**3 + probs**2/a + a/3.0), v)
           #special handling for f == 0 poissonPicture terms using quadratic rounding of function with minimum: max(0,(a-p))^2/(2a) + p

        dprobs_factor_pos = (countVecMx / pos_probs - totalCntVec[None,:])
        dprobs_factor_neg = S + 2*S2*(probs - min_p)
        dprobs_factor_zerofreq = -totalCntVec[None,:] * _np.where( probs >= a, 1.0, (-1.0/a**2)*probs**2 + 2*probs/a)
        dprobs_factor = _np.where( probs < min_p, dprobs_factor_neg, dprobs_factor_pos)
        dprobs_factor = _np.where( countVecMx == 0, dprobs_factor_zerofreq, dprobs_factor )
        jac = dprobs * dprobs_factor[:,:,None] # (K,M,N) * (K,M,1)   (N = dim of vectorized gateset)


    else: #(the non-poisson picture requires that the probabilities of the spam labels for a given string are constrained to sum to 1)
        S = countVecMx / min_p              # slope term that is derivative of logl at min_p
        S2 = -0.5 * countVecMx / (min_p**2) # 2nd derivative of logl term at min_p
        v = countVecMx * _np.log(pos_probs) # dims K x M (K = nSpamLabels, M = nGateStrings)
        v = _np.minimum(v,0)  #remove small positive elements due to roundoff error (above expression *cannot* really be positive)
        v = _np.where( probs < min_p, v + S*(probs - min_p) + S2*(probs - min_p)**2, v) #quadratic extrapolation of logl at min_p for probabilities < min_p
        v = _np.where( countVecMx == 0, 0.0, v)

        dprobs_factor_pos = countVecMx / pos_probs
        dprobs_factor_neg = S + 2*S2*(probs - min_p)
        dprobs_factor = _np.where( probs < min_p, dprobs_factor_neg, dprobs_factor_pos)
        dprobs_factor = _np.where( countVecMx == 0, 0.0, dprobs_factor )
        jac = dprobs * dprobs_factor[:,:,None] # (K,M,N) * (K,M,1)   (N = dim of vectorized gateset)

    # jac[iSpamLabel,iGateString,iGateSetParam] contains all d(logl)/d(gatesetParam) contributions
    return _np.sum(jac, axis=(0,1)) # sum over spam label and gate string dimensions


def logl_hessian(gateset, dataset, gatestring_list=None, minProbClip=1e-6,
                 probClipInterval=(-1e6,1e6), radius=1e-4, poissonPicture=True,
                 check=False, comm=None, memLimit=None, verbosity=0):
    """
    The hessian of the log-likelihood function.

    Parameters
    ----------
    gateset : GateSet
        Gateset of parameterized gates (including SPAM)

    dataset : DataSet
        Probability data

    gatestring_list : list of (tuples or GateStrings), optional
        Each element specifies a gate string to include in the log-likelihood
        sum.  Default value of None implies all the gate strings in dataset
        should be used.

    minProbClip : float, optional
        The minimum probability treated normally in the evaluation of the log-likelihood.
        A penalty function replaces the true log-likelihood for probabilities that lie
        below this threshold so that the log-likelihood never becomes undefined (which improves
        optimizer performance).

    probClipInterval : 2-tuple or None, optional
        (min,max) values used to clip the probabilities predicted by
        gatesets during MLEGST's search for an optimal gateset (if not None).
        if None, no clipping is performed.

    radius : float, optional
        Specifies the severity of rounding used to "patch" the zero-frequency
        terms of the log-likelihood.

    poissonPicture : boolean, optional
        Whether the Poisson-picutre log-likelihood should be differentiated.

    check : boolean, optional
        If True, perform extra checks within code to verify correctness.  Used
        for testing, and runs much slower when True.

    comm : mpi4py.MPI.Comm, optional
        When not None, an MPI communicator for distributing the computation
        across multiple processors.

    memLimit : int, optional
        A rough memory limit in bytes which restricts the amount of intermediate
        values that are computed and stored.

    verbosity : int, optional
        How much detail to print to stdout.


    Returns
    -------
    numpy array
      array of shape (M,M), where M is the length of the vectorized gateset.
    """

    nP = gateset.num_params()

    if gatestring_list is None:
        gatestring_list = list(dataset.keys())

    spamLabels = gateset.get_spam_labels() #fixes the ordering of the spam labels
    spam_lbl_rows = { sl:i for (i,sl) in enumerate(spamLabels) }
    
    #  Estimate & check persistent memory (from allocs directly below)
    C = 1.0/1024.0**3; nP = gateset.num_params()
    persistentMem = 8*nP**2 # in bytes
    if memLimit is not None and memLimit < persistentMem:
        raise MemoryError("HLogL Memory limit (%g GB) is " % (memLimit*C) +
                          "< memory required to hold final results (%g GB)"
                          % (persistentMem*C))

    #  Allocate persistent memory
    final_hessian = _np.zeros( (nP,nP), 'd')

    #  Estimate & check intermediate memory
    #  - figure out how many row & column partitions are needed
    #    to fit computation within available memory (and use all cpus)
    mlim = None if (memLimit is None) else memLimit-persistentMem
    evalTree, blkSize1, blkSize2 = gateset.bulk_evaltree_from_resources(
        gatestring_list, comm, mlim, "deriv", ['bulk_hprobs_by_block'],
        verbosity)
    
    rowParts = int(round(nP / blkSize1)) if (blkSize1 is not None) else 1
    colParts = int(round(nP / blkSize2)) if (blkSize2 is not None) else 1

    a = radius # parameterizes "roundness" of f == 0 terms
    min_p = minProbClip

    if poissonPicture:
        #NOTE: hessian_from_hprobs MAY modify hprobs and dprobs12 (to save mem)
        def hessian_from_hprobs(hprobs, dprobs12, cntVecMx, totalCntVec, pos_probs):
            # Notation:  (K=#spam, M=#strings, N=#wrtParams1, N'=#wrtParams2 )
            totCnts = totalCntVec[None,:]  #shorthand (just a view)
            S = cntVecMx / min_p - totCnts # slope term that is derivative of logl at min_p
            S2 = -0.5 * cntVecMx / (min_p**2)          # 2nd derivative of logl term at min_p

            #hprobs_pos  = (-cntVecMx / pos_probs**2)[:,:,None,None] * dprobs12   # (K,M,1,1) * (K,M,N,N')
            #hprobs_pos += (cntVecMx / pos_probs - totalCntVec[None,:])[:,:,None,None] * hprobs  # (K,M,1,1) * (K,M,N,N')
            #hprobs_neg  = (2*S2)[:,:,None,None] * dprobs12 + (S + 2*S2*(probs - min_p))[:,:,None,None] * hprobs # (K,M,1,1) * (K,M,N,N')
            #hprobs_zerofreq = _np.where( (probs >= a)[:,:,None,None],
            #                             -totalCntVec[None,:,None,None] * hprobs,
            #                             (-totalCntVec[None,:] * ( (-2.0/a**2)*probs + 2.0/a))[:,:,None,None] * dprobs12
            #                             - (totalCntVec[None,:] * ((-1.0/a**2)*probs**2 + 2*probs/a))[:,:,None,None] * hprobs )
            #hessian = _np.where( (probs < min_p)[:,:,None,None], hprobs_neg, hprobs_pos)
            #hessian = _np.where( (cntVecMx == 0)[:,:,None,None], hprobs_zerofreq, hessian) # (K,M,N,N')
            
            #Accomplish the same thing as the above commented-out lines, 
            # but with more memory effiency:
            dprobs12_coeffs = \
                _np.where(probs < min_p, 2*S2, -cntVecMx / pos_probs**2)
            zfc = _np.where(probs >= a, 0.0, -totCnts*((-2.0/a**2)*probs+2.0/a))
            dprobs12_coeffs = _np.where(cntVecMx == 0, zfc, dprobs12_coeffs)

            hprobs_coeffs = \
                _np.where(probs < min_p, S + 2*S2*(probs - min_p),
                          cntVecMx / pos_probs - totCnts)
            zfc = _np.where(probs >= a, -totCnts, 
                            -totCnts * ((-1.0/a**2)*probs**2 + 2*probs/a))
            hprobs_coeffs = _np.where(cntVecMx == 0, zfc, hprobs_coeffs)

              # hessian = hprobs_coeffs * hprobs + dprobs12_coeff * dprobs12
              #  but re-using dprobs12 and hprobs memory (which is overwritten!)
            hprobs *= hprobs_coeffs[:,:,None,None]
            dprobs12 *= dprobs12_coeffs[:,:,None,None]
            hessian = dprobs12; hessian += hprobs

            # hessian[iSpamLabel,iGateString,iGateSetParam1,iGateSetParams2] contains all
            #  d2(logl)/d(gatesetParam1)d(gatesetParam2) contributions
            return _np.sum(hessian, axis=(0,1))
              # sum over spam label and gate string dimensions (gate strings in evalSubTree)
              # adds current subtree contribution for (N,N')-sized block of Hessian


    else:

        #(the non-poisson picture requires that the probabilities of the spam labels for a given string are constrained to sum to 1)
        #NOTE: hessian_from_hprobs MAY modify hprobs and dprobs12 (to save mem)
        def hessian_from_hprobs(hprobs, dprobs12, cntVecMx, totalCntVec, pos_probs):
            S = cntVecMx / min_p # slope term that is derivative of logl at min_p
            S2 = -0.5 * cntVecMx / (min_p**2) # 2nd derivative of logl term at min_p

            #hprobs_pos  = (-cntVecMx / pos_probs**2)[:,:,None,None] * dprobs12   # (K,M,1,1) * (K,M,N,N')
            #hprobs_pos += (cntVecMx / pos_probs)[:,:,None,None] * hprobs  # (K,M,1,1) * (K,M,N,N')
            #hprobs_neg  = (2*S2)[:,:,None,None] * dprobs12 + (S + 2*S2*(probs - min_p))[:,:,None,None] * hprobs # (K,M,1,1) * (K,M,N,N')
            #hessian = _np.where( (probs < min_p)[:,:,None,None], hprobs_neg, hprobs_pos)
            #hessian = _np.where( (cntVecMx == 0)[:,:,None,None], 0.0, hessian) # (K,M,N,N')

            #Accomplish the same thing as the above commented-out lines, 
            # but with more memory effiency:
            dprobs12_coeffs = \
                _np.where(probs < min_p, 2*S2, -cntVecMx / pos_probs**2)
            dprobs12_coeffs = _np.where(cntVecMx == 0, 0.0, dprobs12_coeffs)

            hprobs_coeffs = \
                _np.where(probs < min_p, S + 2*S2*(probs - min_p),
                          cntVecMx / pos_probs)
            hprobs_coeffs = _np.where(cntVecMx == 0, 0.0, hprobs_coeffs)

              # hessian = hprobs_coeffs * hprobs + dprobs12_coeff * dprobs12
              #  but re-using dprobs12 and hprobs memory (which is overwritten!)
            hprobs *= hprobs_coeffs[:,:,None,None]
            dprobs12 *= dprobs12_coeffs[:,:,None,None]
            hessian = dprobs12; hessian += hprobs

            return _np.sum(hessian, axis=(0,1)) #see comments as above


    #Note - we could in the future use comm to distribute over
    # subtrees here.  We currently don't because we parallelize
    # over columns and it seems that in almost all cases of
    # interest there will be more hessian columns than processors,
    # so adding the additional ability to parallelize over
    # subtrees would just add unnecessary complication.

    #get distribution across subtrees (groups if needed)
    subtrees = evalTree.get_sub_trees()
    mySubTreeIndices, subTreeOwners, mySubComm = evalTree.distribute(comm)

    #  Allocate memory (alloc max required & take views)
    maxNumGatestrings = max([subtrees[i].num_final_strings() for i in mySubTreeIndices])
    cntVecMx_mem = _np.empty( (len(spamLabels),maxNumGatestrings),'d')
    probs_mem  = _np.empty( (len(spamLabels),maxNumGatestrings), 'd' )

    #DEBUG
    #import time
    #import sys
    #tStart = time.time()

    #Loop over subtrees
    for iSubTree in mySubTreeIndices:
        evalSubTree = subtrees[iSubTree]
        sub_nGateStrings = evalSubTree.num_final_strings()

        #  Create views into pre-allocated memory
        cntVecMx = cntVecMx_mem[:,0:sub_nGateStrings]
        probs  =  probs_mem[:,0:sub_nGateStrings]

        # Fill cntVecMx, totalCntVec
        fill_count_vecs(cntVecMx,spam_lbl_rows,dataset,
                            evalSubTree.generate_gatestring_list())
        totalCntVec = _np.sum(cntVecMx, axis=0)

        #compute pos_probs separately
        gateset.bulk_fill_probs(probs, spam_lbl_rows, evalSubTree,
                                clipTo=probClipInterval, check=check,
                                comm=mySubComm)
        pos_probs = _np.where(probs < min_p, min_p, probs)

        nCols = gateset.num_params()
        blocks1 = _mpit.slice_up_range(nCols, rowParts)
        blocks2 = _mpit.slice_up_range(nCols, colParts)
        sliceTupList_all = list(_itertools.product(blocks1,blocks2))
        #cull out lower triangle blocks, which have no overlap with
        # the upper triangle of the hessian
        sliceTupList = [ (slc1,slc2) for slc1,slc2 in sliceTupList_all
                         if slc1.start <= slc2.stop ]

        loc_iBlks, blkOwners, blkComm = \
            _mpit.distribute_indices(list(range(len(sliceTupList))), mySubComm)
        mySliceTupList = [ sliceTupList[i] for i in loc_iBlks ]
       
        subtree_hessian = _np.zeros( (nP,nP), 'd')

        #k,kmax = 0,len(mySliceTupList) #DEBUG
        for (slice1,slice2,hprobs,dprobs12) in gateset.bulk_hprobs_by_block(
            spam_lbl_rows, evalSubTree, mySliceTupList, True, blkComm):

            #DEBUG
            #iSub = mySubTreeIndices.index(iSubTree)
            #print("DEBUG: rank%d: %gs: block %d/%d, sub-tree %d/%d, sub-tree-len = %d"
            #          % (comm.Get_rank(),time.time()-tStart,k,kmax,iSub,
            #             len(mySubTreeIndices), len(evalSubTree)))            
            #sys.stdout.flush(); k += 1

            subtree_hessian[slice1,slice2] = \
                hessian_from_hprobs(hprobs, dprobs12, cntVecMx,
                                        totalCntVec, pos_probs)
                #NOTE: hessian_from_hprobs MAY modify hprobs and dprobs12

        #Gather columns from different procs and add to running final hessian
        #_mpit.gather_slices_by_owner(slicesIOwn, subtree_hessian, (0,1), mySubComm)
        _mpit.gather_slices(sliceTupList, blkOwners, subtree_hessian, (0,1), mySubComm)
        final_hessian += subtree_hessian

    #gather (add together) final_hessians from different processors
    if comm is not None and len(set(subTreeOwners.values())) > 1:
        if comm.Get_rank() not in subTreeOwners.values(): 
            # this proc is not the "owner" of its subtrees and should not send a contribution to the sum
            final_hessian[:,:] = 0.0 #zero out hessian so it won't contribute
        final_hessian = comm.allreduce(final_hessian)
        
    #copy upper triangle to lower triangle (we only compute upper)
    for i in range(final_hessian.shape[0]):
        for j in range(i+1,final_hessian.shape[1]):
            final_hessian[j,i] = final_hessian[i,j]

    return final_hessian # (N,N)


def logl_max(dataset, gatestring_list=None, countVecMx=None, poissonPicture=True, check=False):
    """
    The maximum log-likelihood possible for a DataSet.  That is, the
    log-likelihood obtained by a maximal model that can fit perfectly
    the probability of each gate string.

    Parameters
    ----------
    dataset : DataSet
        the data set to use.

    gatestring_list : list of (tuples or GateStrings), optional
        Each element specifies a gate string to include in the max-log-likelihood
        sum.  Default value of None implies all the gate strings in dataset should
        be used.

    countVecMx : numpy array, optional
        Two-dimensional numpy array whose rows correspond to the data set's spam
        labels (i.e. dataset.get_spam_labels()).  Each row is  contains the
        dataset counts for that spam label for each gate string in gatestring_list.
        Use fill_count_vecs(...) to generate this quantity when it is useful elsewhere
        (e.g. for logl(...) calls).

    poissonPicture : boolean, optional
        Whether the Poisson-picture maximum log-likelihood should be returned.

    check : boolean, optional
        Whether additional check is performed which computes the max logl another
        way an compares to the faster method.

    Returns
    -------
    float
    """

    if gatestring_list is None:
        gatestring_list = list(dataset.keys())

    if countVecMx is None:
        spamLabels = dataset.get_spam_labels()
        spam_lbl_rows = { sl:i for (i,sl) in enumerate(spamLabels) }
        countVecMx = _np.empty( (len(spamLabels),len(gatestring_list)), 'd' )
        fill_count_vecs(countVecMx, spam_lbl_rows, dataset, gatestring_list)

    totalCntVec = _np.sum(countVecMx, axis=0)
    freqs = countVecMx / totalCntVec[None,:]
    freqs_nozeros = _np.where(countVecMx == 0, 1.0, freqs) # set zero freqs to 1.0 so np.log doesn't complain

    if poissonPicture:
        maxLogLTerms = countVecMx * ( _np.log(freqs_nozeros) - 1.0 )
    else:
        maxLogLTerms = countVecMx * _np.log(freqs_nozeros)

    maxLogLTerms[ countVecMx == 0 ] = 0.0 # set 0 * log(0) terms explicitly to zero since numpy doesn't know this limiting behavior

    # maxLogLTerms[iSpamLabel,iGateString] contains all logl-upper-bound contributions
    maxLogL = _np.sum(maxLogLTerms) # sum over *all* dimensions

    if check:
        L = 0
        for gateString in gatestring_list:
            dsRow = dataset[gateString]
            N = dsRow.total() #sum of counts for all outcomes (all spam labels)
            for n in list(dsRow.values()):
                f = n / N
                if f < TOL and n == 0: continue # 0 * log(0) == 0
                if poissonPicture:
                    L += n * _np.log(f) - N * f
                else:
                    L += n * _np.log(f)
        if not _np.isclose(maxLogL,L):
            _warnings.warn("Log-likelihood upper bound mismatch: %g != %g (diff=%g)" % \
                               (maxLogL, L, maxLogL-L))

    return maxLogL


def forbidden_prob(gateset, dataset):
    """
    Compute the sum of the out-of-range probabilities
    generated by gateset, using only those gate strings
    contained in dataset.  Non-zero value indicates
    that gateset is not in XP for the supplied dataset.

    Parameters
    ----------
    gateset : GateSet
        gate set to generate probabilities.

    dataset : DataSet
        data set to obtain gate strings.  Dataset counts are
        used to check for zero or all counts being under a
        single spam label, in which case out-of-bounds probabilities
        are ignored because they contribute zero to the logl sum.

    Returns
    -------
    float
        sum of the out-of-range probabilities.
    """
    forbidden_prob = 0

    for gs,dsRow in dataset.iteritems():
        probs = gateset.probs(gs)
        for (spamLabel,p) in probs.items():
            if p < TOL:
                if round(dsRow[spamLabel]) == 0: continue #contributes zero to the sum
                else: forbidden_prob += abs(TOL-p) + TOL
            elif p > 1-TOL:
                if round(dsRow[spamLabel]) == dsRow.total(): continue #contributes zero to the sum
                else: forbidden_prob += abs(p-(1-TOL)) + TOL


    return forbidden_prob

def prep_penalty(rhoVec):
    """
    Penalty assigned to a state preparation (rho) vector rhoVec.  State
      preparation density matrices must be positive semidefinite
      and trace == 1.  A positive return value indicates an
      these criteria are not met and the rho-vector is invalid.

    Parameters
    ----------
    rhoVec : numpy array
        rho vector array of shape (N,1) for some N.

    Returns
    -------
    float
    """
    # rhoVec must be positive semidefinite and trace = 1
    rhoMx = _bt.gmvec_to_stdmx(_np.asarray(rhoVec))
    evals = _np.linalg.eigvals( rhoMx )  #could use eigvalsh, but wary of this since eigh can be wrong...
    sumOfNeg = sum( [ -ev.real for ev in evals if ev.real < 0 ] )
    nQubits = _np.log2(len(rhoVec)) / 2
    tracePenalty = abs(rhoVec[0,0]-(1.0/_np.sqrt(2))**nQubits) # tensor of n I(2x2)/sqrt(2) has trace sqrt(2)**n
    #print "Sum of neg = ",sumOfNeg  #DEBUG
    #print "Trace Penalty = ",tracePenalty  #DEBUG
    return sumOfNeg +  tracePenalty

def effect_penalty(EVec):
    """
    Penalty assigned to a POVM effect vector EVec. Effects
      must have eigenvalues between 0 and 1.  A positive return
      value indicates this criterion is not met and the E-vector
      is invalid.

    Parameters
    ----------
    EVec : numpy array
         effect vector array of shape (N,1) for some N.

    Returns
    -------
    float
    """
    # EVec must have eigenvalues between 0 and 1
    EMx = _bt.gmvec_to_stdmx(_np.asarray(EVec))
    evals = _np.linalg.eigvals( EMx )  #could use eigvalsh, but wary of this since eigh can be wrong...
    sumOfPen = 0
    for ev in evals:
        if ev.real < 0: sumOfPen += -ev.real
        if ev.real > 1: sumOfPen += ev.real-1.0
    return sumOfPen

def cptp_penalty(gateset, include_spam_penalty=True):
    """
    The sum of all negative Choi matrix eigenvalues, and
      if include_spam_penalty is True, the rho-vector and
      E-vector penalties of gateset.  A non-zero value
      indicates that the gateset is not CPTP.

    Parameters
    ----------
    gateset : GateSet
        the gate set to compute CPTP penalty for.

    include_spam_penalty : bool, optional
        if True, also test gateset for invalid SPAM
        operation(s) and return sum of CPTP penalty
        with rhoVecPenlaty(...) and effect_penalty(...)
        for each rho and E vector.

    Returns
    -------
    float
        CPTP penalty (possibly with added spam penalty).
    """
    ret = _jam.sum_of_negative_choi_evals(gateset)
    if include_spam_penalty:
        ret += sum([ prep_penalty(r) for r in list(gateset.preps.values()) ])
        ret += sum([ effect_penalty(e) for e in list(gateset.effects.values()) ])
    return ret


def two_delta_loglfn(N, p, f, minProbClip=1e-6, poissonPicture=True):
    """
    Term of the 2*[log(L)-upper-bound - log(L)] sum corresponding
     to a single gate string and spam label.

    Parameters
    ----------
    N : float or numpy array
        Number of samples.

    p : float or numpy array
        Probability of 1st outcome (typically computed).

    f : float or numpy array
        Frequency of 1st outcome (typically observed).

    minProbClip : float, optional
        Minimum probability clip point to avoid evaluating
        log(number <= zero)

    poissonPicture : boolean, optional
        Whether the log-likelihood-in-the-Poisson-picture terms should be included
        in the returned logl value.

    Returns
    -------
    float or numpy array
    """
    #TODO: change this function to handle nan's in the inputs without warnings, since
    # fiducial pair reduction may pass inputs with nan's legitimately and the desired
    # behavior is to just let the nan's pass through to nan's in the output.
    cp  = _np.clip(p,minProbClip,1e10) #effectively no upper bound
    
    nan_indices = _np.isnan(f) # get indices of invalid entries
    if not _np.isscalar(f): f[nan_indices] = 0.0 
      #set nan's to zero to avoid RuntimeWarnings (invalid value)

    zf  = _np.where(f < 1e-10, 0.0, f) #set zero-freqs to zero
    nzf = _np.where(f < 1e-10, 1.0, f) #set zero-freqs to one -- together
                                       # w/above line makes 0 * log(0) == 0    
    if not _np.isscalar(f): 
        zf[nan_indices] = _np.nan  #set nan indices back to nan
        nzf[nan_indices] = _np.nan #set nan indices back to nan

    if poissonPicture:
        return 2 * (N * zf * _np.log(nzf/cp) - N * (f-cp))
    else:
        return 2 * N * zf * _np.log(nzf/cp)






##############################################################################################
#   FUNCTIONS FOR HESSIAN ANALYSIS (which take derivatives of the log(likelihood) function)  #
##############################################################################################


#def dlogl_analytic(gateset, dataset):
#    nP = gateset.num_params()
#    result = _np.zeros([1,nP])
#    dPmx = dpr_plus(gateset, [gateString for gateString in dataset])
#
#    for (k,d) in enumerate(dataset.values()):
#        p = gateset.PrPlus(d.gateString)
#        if _np.fabs(p) < TOL and round(d.nPlus) == 0: continue
#        if _np.fabs(p - 1) < TOL and round(d.nMinus) == 0: continue
#
#        for i in range(nP):
#            #pre = ((1-p)*d.nPlus - p*d.nMinus) / (p*(1-p))
#            #print "%d: Pre(%s) = " % (i,d.gateString), pre, "  (p = %g, np = %g)" % (p, d.nPlus)
#            result[0,i] += ((1-p)*d.nPlus - p*d.nMinus) / (p*(1-p)) * dPmx[i,k]
#
#    return result
#
#
#def dlogl_finite_diff(gateset, dataset):
#    return numerical_deriv(logl, gateset, dataset, 1)
#
#def logl_hessian_finite_diff(gateset, dataset):
#    return numerical_deriv(dlogl_finite_diff, gateset, dataset, gateset.num_params())
#
#def logl_hessian_at_ml(gateset, gatestrings, nSamples):
#    return nSamples * logl_hessian_at_ML_per_sample(gateset, gatestrings)
#
#def logl_hessian_at_ML_per_sample(gateset, gatestrings):
#    nP = gateset.num_params()
#    result = _np.zeros([nP,nP])
#
#    dPmx = dpr_plus(gateset, gatestrings)
#
#    for (k,s) in enumerate(gatestrings):
#        p = gateset.PrPlus(s)
#        if _np.fabs(p) < TOL: continue
#        if _np.fabs(p - 1) < TOL: continue
#        for i in range(nP):
#            for j in range(nP):
#                result[i,j] += -1.0/(p*(1-p)) * dPmx[i,k] * dPmx[j,k]
#
#    return result
#
#
#
#def dpr_plus(gateset, gatestrings):
#    DELTA = 1e-7
#    nP = gateset.num_params()
#    nGateStrings = len(gatestrings)
#    result = _np.zeros([nP,nGateStrings])
#
#    for (j,s) in enumerate(gatestrings):
#        fMid = gateset.PrPlus(s)
#
#        for i in range(nP):
#            gs = gateset.copy()
#            gs.add_to_param(i,DELTA)
#            fRight = gs.PrPlus(s)
#            gs.add_to_param(i,-2*DELTA)
#            fLeft = gs.PrPlus(s)
#
#            if fRight is None and fLeft is None:
#                raise ValueError("Cannot take derivative - both sides are out of bounds!")
#            if fRight is None:
#                dP = (fMid - fLeft) / DELTA
#            elif fLeft is None:
#                dP = (fRight - fMid) / DELTA
#            else:
#                dP = (fRight - fLeft) / (2*DELTA)
#
#            result[i,j] = dP
#
#    return result
#
#
#def numerical_deriv(fnToDifferentiate, gateset, dataset, resultLen):
#    DELTA = 1e-6
#    nP = gateset.num_params()
#    result = _np.zeros([resultLen,nP])
#
#    fMid = fnToDifferentiate(gateset, dataset)
#    if fMid is None: return None
#
#    for i in range(nP):
#        gs = gateset.copy()
#        gs.add_to_param(i,DELTA)
#        fRight = fnToDifferentiate(gs, dataset)
#
#        gs = gateset.copy()
#        gs.add_to_param(i,-DELTA)
#        fLeft = fnToDifferentiate(gs, dataset)
#
#        #print "DEBUG: %d: l,m,r = " % i,(fLeft,fMid,fRight)
#        if fRight is None and fLeft is None:
#            raise ValueError("numerical_deriv cannot take derivative - both sides are out of bounds!")
#
#        if fRight is None:
#            df = (fMid - fLeft) / DELTA
#        elif fLeft is None:
#            df = (fRight - fMid) / DELTA
#        else:
#            df = (fRight - fLeft) / (2*DELTA)
#
#        #print "DEBUG: df(%d) = " % i,df
#        result[:,i] = _np.transpose(df)
#
#    return result



from __future__ import division, print_function, absolute_import, unicode_literals
#*****************************************************************
#    pyGSTi 0.9:  Copyright 2015 Sandia Corporation
#    This Software is released under the GPL license detailed
#    in the file "license.txt" in the top-level pyGSTi directory
#*****************************************************************
""" Defines the MultiDataSet class and supporting classes and functions """

import numpy as _np
import pickle as _pickle
from collections import OrderedDict as _OrderedDict

from .dataset import DataSet as _DataSet
from . import gatestring as _gs


class MultiDataSet_KeyValIterator(object):
    """ Iterator class for datasetName,DataSet pairs of a MultiDataSet """
    def __init__(self, multidataset):
        self.multidataset = multidataset
        self.countsDictIter = multidataset.countsDict.__iter__()

    def __iter__(self):
        return self

    def __next__(self):
        datasetName = next(self.countsDictIter)
        return datasetName, _DataSet(self.multidataset.countsDict[datasetName], 
                                     gateStringIndices=self.multidataset.gsIndex,
                                     spamLabelIndices=self.multidataset.slIndex,
                                     collisionAction=self.multidataset.collisionActions[datasetName],
                                     bStatic=True)

    next = __next__


class MultiDataSet_ValIterator(object):
    """ Iterator class for DataSets of a MultiDataSet """
    def __init__(self, multidataset):
        self.multidataset = multidataset
        self.countsDictIter = multidataset.countsDict.__iter__()

    def __iter__(self):
        return self

    def __next__(self):
        datasetName = next(self.countsDictIter)
        return _DataSet(self.multidataset.countsDict[datasetName],
                        gateStringIndices=self.multidataset.gsIndex,
                        spamLabelIndices=self.multidataset.slIndex, bStatic=True,
                        collisionAction=self.multidataset.collisionActions[datasetName],)

    next = __next__

class MultiDataSet(object):
    """
    The MultiDataSet class allows for the combined access and storage of
    several static DataSets that contain the same gate strings (in the same order).
    It is designed to behave similarly to a dictionary of DataSets, so that
    a DataSet is obtained by (Note that datasetName may be a tuple):

    dataset = multiDataset[datasetName]
    """

    def __init__(self, countsDict=None,
                 gateStrings=None, gateStringIndices=None,
                 spamLabels=None, spamLabelIndices=None,
                 fileToLoadFrom=None, collisionActions=None,
                 comment=None):
        """
        Initialize a MultiDataSet.

        Parameters
        ----------
        countsDict : ordered dictionary, optional
          Keys specify dataset names.  Values are 2D numpy arrays which specify counts. Rows of the arrays
          correspond to gate strings and columns to spam labels.

        gateStrings : list of (tuples or GateStrings), optional
          Each element is a tuple of gate labels or a GateString object.  Indices for these strings
          are assumed to ascend from 0.  These indices must correspond to rows/elements of counts (above).
          Only specify this argument OR gateStringIndices, not both.

        gateStringIndices : ordered dictionary, optional
          An OrderedDict with keys equal to gate strings (tuples of gate labels) and values equal to
          integer indices associating a row/element of counts with the gate string.  Only
          specify this argument OR gateStrings, not both.

        spamLabels : list of strings, optional
          Specifies the set of spam labels for the DataSet.  Indices for the spam labels
          are assumed to ascend from 0, starting with the first element of this list.  These
          indices will index columns of the counts array/list.  Only specify this argument
          OR spamLabelIndices, not both.

        spamLabelIndices : ordered dictionary, optional
          An OrderedDict with keys equal to spam labels (strings) and value  equal to
          integer indices associating a spam label with a column of counts.  Only
          specify this argument OR spamLabels, not both.

        fileToLoadFrom : string or file object, optional
          Specify this argument and no others to create a MultiDataSet by loading
          from a file (just like using the load(...) function).

        collisionActions : dictionary, optional
            Specifies how duplicate gate sequences should be handled for the data
            sets specified by `countsDict`.  Keys must match those of `countsDict`
            and values are "aggregate" or "keepseparate".  See documentation for
            `DataSet`.  If None, then "aggregate" is used for all sets by default.

        comment : string, optional
            A user-specified comment string that gets carried around with the 
            data.  A common use for this field is to attache to the data details
            regarding its collection.

        Returns
        -------
        MultiDataSet
           a new multi data set object.
        """

        #Optionally load from a file
        if fileToLoadFrom is not None:
            assert(countsDict is None and gateStrings is None and gateStringIndices is None and spamLabels is None and spamLabelIndices is None)
            self.load(fileToLoadFrom)
            return

        # self.gsIndex  :  Ordered dictionary where keys = gate strings (tuples), values = integer indices into counts
        if gateStringIndices is not None:
            self.gsIndex = gateStringIndices
        elif gateStrings is not None:
            dictData = [ (gs if isinstance(gs,_gs.GateString) else _gs.GateString(gs),i) \
                           for (i,gs) in enumerate(gateStrings) ] #convert to GateStrings if necessary
            self.gsIndex = _OrderedDict( dictData )
        else:
            self.gsIndex = None

        # self.slIndex  :  Ordered dictionary where keys = spam labels (strings), values = integer indices into counts
        if spamLabelIndices is not None:
            self.slIndex = spamLabelIndices
        elif spamLabels is not None:
            self.slIndex = _OrderedDict( [(sl,i) for (i,sl) in enumerate(spamLabels) ] )
        else:
            self.slIndex = None

        if self.gsIndex:  #Note: tests if not none and nonempty
            assert( min(self.gsIndex.values()) >= 0)
        if self.slIndex:  #Note: tests if not none and nonempty
            assert( min(self.slIndex.values()) >= 0)

        # self.countsDict : a dictionary of 2D numpy arrays, each corresponding to a DataSet.  Rows = gate strings, Cols = spam labels
        #                   ( keys = dataset names, values = 2D counts array of corresponding dataset )
        if countsDict is not None:
            self.countsDict = _OrderedDict( [ (name,counts) for name,counts in countsDict.items() ] ) #copy OrderedDict but share counts arrays
            if collisionActions is None: collisionActions = {} #allow None to function as an empty dict
            self.collisionActions = _OrderedDict( [ (name,collisionActions.get(name,"aggregate"))
                                                    for name in self.countsDict.keys() ] )

            if self.gsIndex:  #Note: tests if not none and nonempty
                #minIndex = min(self.gsIndex.values())
                maxIndex = max(self.gsIndex.values())
                for _,counts in self.countsDict.items():
                    assert( counts.shape[0] > maxIndex and counts.shape[1] == len(self.slIndex) )
        else:
            self.countsDict = _OrderedDict()
            self.collisionActions = _OrderedDict()

        # comment
        self.comment = comment


    def get_spam_labels(self):
        """
        Get the spam labels of this MultiDataSet.

        Returns
        -------
        list of strings
          A list where each element is a spam label.
          Returns None when the MultiDataSet is not
          yet initialized with any data or spam labels.
        """
        if self.slIndex is not None:
            return list(self.slIndex.keys())
        else: return None


    def __iter__(self):
        return self.countsDict.__iter__() #iterator over dataset names

    def __len__(self):
        return len(self.countsDict)

    def __getitem__(self, datasetName):  #return a static DataSet
        return _DataSet(self.countsDict[datasetName],
                        gateStringIndices=self.gsIndex,
                        spamLabelIndices=self.slIndex, bStatic=True,
                        collisionAction=self.collisionActions[datasetName])

    def __setitem__(self, datasetName, dataset):
        self.add_dataset(datasetName, dataset)

    def __contains__(self, datasetName):
        return datasetName in self.countsDict

    def keys(self):
        """ Returns a list of the keys (dataset names) of this MultiDataSet """
        return list(self.countsDict.keys())

    def has_key(self, datasetName):
        """ Test whether this MultiDataSet contains a given dataset name """
        return datasetName in self.countsDict

    def iteritems(self):
        """ Iterator over (dataset name, DataSet) pairs """
        return MultiDataSet_KeyValIterator(self)

    def itervalues(self):
        """ Iterator over DataSets corresponding to each dataset name """
        return MultiDataSet_ValIterator(self)

    def get_datasets_sum(self, *datasetNames):
        """
        Generate a new DataSet by combining the counts of multiple member Datasets.

        Parameters
        ----------
        datasetNames : one or more dataset names.

        Returns
        -------
        DataSet
            a single DataSet containing the summed counts of each of the datasets
            named by the parameters.
        """
        summedCounts = None
        if len(datasetNames) == 0: raise ValueError("Must specify at least one dataset name")
        for datasetName in datasetNames:
            if datasetName not in self:
                raise ValueError("No dataset with the name '%s' exists" % datasetName)

            if summedCounts is None:
                summedCounts = self.countsDict[datasetName].copy()
            else:
                summedCounts += self.countsDict[datasetName]

        return _DataSet(summedCounts, gateStringIndices=self.gsIndex,
                        spamLabelIndices=self.slIndex, bStatic=True)
                        #leave collisionAction as default "aggregate"

    def add_dataset(self, datasetName, dataset):
        """
        Add a DataSet to this MultiDataSet.  The dataset
        must be static and conform with the gate strings passed
        upon construction or those inherited from the first
        dataset added.

        Parameters
        ----------
        datasetName : string
            The name to give the added dataset (i.e. the key the new
            data set will be referenced by).

        dataset : DataSet
            The data set to add.
        """

        #first test if dataset is compatible
        if not dataset.bStatic:
            raise ValueError("Cannot add dataset: only static DataSets can be added to a MultiDataSet")
        if self.gsIndex is not None and dataset.gsIndex != self.gsIndex:
            raise ValueError("Cannot add dataset: gate strings and/or their indices do not match")
        if self.slIndex is not None and dataset.slIndex != self.slIndex:
            raise ValueError("Cannot add dataset: spam labels and/or their indices do not match")

        if self.gsIndex:  #Note: tests if not none and nonempty
            maxIndex = max(self.gsIndex.values())
            assert( dataset.counts.shape[0] > maxIndex and dataset.counts.shape[1] == len(self.slIndex) )

        self.countsDict[datasetName] = dataset.counts
        self.collisionActions[datasetName] = dataset.collisionAction

        if self.gsIndex is None:
            self.gsIndex = dataset.gsIndex
            if len(self.gsIndex) > 0:
                assert( min(self.gsIndex.values()) >= 0)

        if self.slIndex is None:
            self.slIndex = dataset.slIndex
            if len(self.slIndex) > 0:
                assert( min(self.slIndex.values()) >= 0)


    def add_dataset_counts(self, datasetName, datasetCounts,
                           collisionAction="aggregate"):
        """
        Directly add a full set of counts for a specified dataset.

        Parameters
        ----------
        datasetName : string
            Counts are added for this data set.  This can be a new name, in
            which case this method adds a new data set to the MultiDataSet.

        datasetCounts: numpy array
            A 2D array with rows = gate strings and cols = spam labels, to this
            MultiDataSet.  The shape of dataSetCounts is checked for compatibility.

        collisionAction : {"aggregate", "keepseparate"}
            Specifies how duplicate gate sequences should be handled for this
            data set.  This is applicable only if and when the dataset is copied
            to a non-static one.  "aggregate" adds duplicate-sequence counts,
            whereas "keepseparate" tags duplicate-sequence data with by appending
            a final "#<number>" gate label to the duplicated gate sequence.
        """

        if self.gsIndex:  #Note: tests if not none and nonempty
            maxIndex = max(self.gsIndex.values())
            assert( datasetCounts.shape[0] > maxIndex and datasetCounts.shape[1] == len(self.slIndex) )
        self.countsDict[datasetName] = datasetCounts
        self.collisionActions[datasetName] = collisionAction

    def __str__(self):
        s  = "MultiDataSet containing: %d datasets, each with %d strings\n" % (len(self), len(self.gsIndex) if self.gsIndex is not None else 0)
        s += " Dataset names = " + ", ".join(list(self.keys())) + "\n"
        s += " SPAM labels = " + ", ".join(list(self.slIndex.keys()) if self.slIndex is not None else [])
        if self.gsIndex is not None:
            s += "\nGate strings: \n" + "\n".join( map(str,list(self.gsIndex.keys())) )
        return s + "\n"


    def copy(self):
        """ Make a copy of this MultiDataSet """
        return MultiDataSet(self.countsDict, gateStringIndices=self.gsIndex, spamLabelIndices=self.slIndex,
                            collisionActions=self.collisionActions)


    def __getstate__(self):
        toPickle = { 'gsIndexKeys': list(map(_gs.CompressedGateString, list(self.gsIndex.keys()))) if self.gsIndex else [],
                     'gsIndexVals': list(self.gsIndex.values()) if self.gsIndex else [],
                     'slIndex': self.slIndex,
                     'countsDict': self.countsDict,
                     'collisionActions': self.collisionActions }
        return toPickle

    def __setstate__(self, state_dict):
        gsIndexKeys = [ cgs.expand() for cgs in state_dict['gsIndexKeys'] ]
        self.gsIndex = _OrderedDict( list(zip(gsIndexKeys, state_dict['gsIndexVals'])) )
        self.slIndex = state_dict['slIndex']
        self.countsDict = state_dict['countsDict']
        self.collisionActions = state_dict['collisionActions']

    def save(self, fileOrFilename):
        """
        Save this MultiDataSet to a file.

        Parameters
        ----------
        fileOrFilename : file or string
            Either a filename or a file object.  In the former case, if the
            filename ends in ".gz", the file will be gzip compressed.
        """

        toPickle = { 'gsIndexKeys': list(map(_gs.CompressedGateString, list(self.gsIndex.keys()))) if self.gsIndex else [],
                     'gsIndexVals': list(self.gsIndex.values()) if self.gsIndex else [],
                     'slIndex': self.slIndex,
                     'countsKeys': list(self.countsDict.keys()),
                     'collisionActions' : self.collisionActions,
                     'comment': self.comment }  #Don't pickle countsDict numpy data b/c it's inefficient
        # Compatability for unicode-literal filenames
        bOpen = not (hasattr(fileOrFilename, 'write'))
        if bOpen:
            if fileOrFilename.endswith(".gz"):
                import gzip as _gzip
                f = _gzip.open(fileOrFilename,"wb")
            else:
                f = open(fileOrFilename,"wb")
        else:
            f = fileOrFilename

        _pickle.dump(toPickle,f)
        for _,data in self.countsDict.items():
            _np.save(f, data)
        if bOpen: f.close()


    def load(self, fileOrFilename):
        """
        Load MultiDataSet from a file, clearing any data is contained previously.

        Parameters
        ----------
        fileOrFilename : file or string
            Either a filename or a file object.  In the former case, if the
            filename ends in ".gz", the file will be gzip uncompressed as it is read.
        """
        # Compatability for unicode-literal filenames
        bOpen = not (hasattr(fileOrFilename, 'write'))
        if bOpen:
            if fileOrFilename.endswith(".gz"):
                import gzip as _gzip
                f = _gzip.open(fileOrFilename,"rb")
            else:
                f = open(fileOrFilename,"rb")
        else:
            f = fileOrFilename

        state_dict = _pickle.load(f)
        def expand(x): #to be backward compatible
            assert isinstance(x,_gs.CompressedGateString)
            return x.expand()
            #else:
            #  _warnings.warn("Deprecated dataset format.  Please re-save " +
            #                 "this dataset soon to avoid future incompatibility.")
            #  return _gs.GateString(_gs.CompressedGateString.expand_gate_label_tuple(x))
        gsIndexKeys = [ expand(cgs) for cgs in state_dict['gsIndexKeys'] ]

        #gsIndexKeys = [ cgs.expand() for cgs in state_dict['gsIndexKeys'] ]
        self.gsIndex = _OrderedDict( list(zip(gsIndexKeys, state_dict['gsIndexVals'])) )
        self.slIndex = state_dict['slIndex']
        self.collisionActions = state_dict['collisionActions']
        self.comment = state_dict["comment"]
        self.countsDict = _OrderedDict()
        for key in state_dict['countsKeys']:
            self.countsDict[key] = _np.lib.format.read_array(f) #np.load(f) doesn't play nice with gzip
        if bOpen: f.close()
